#
# Copyright (C) 2016 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

from datetime import timedelta
import functools
from unittest import mock


from neutron_lib.db import api as db_api

from networking_odl.common import constants as odl_const
from networking_odl.db import db
from networking_odl.db import models
from networking_odl.tests.unit import test_base_db


def in_session(fn):
    @functools.wraps(fn)
    def wrapper(self, *args, **kwargs):
        with db_api.CONTEXT_WRITER.using(self.db_context):
            return fn(self, *args, **kwargs)
    return wrapper


class DbTestCase(test_base_db.ODLBaseDbTestCase):
    def setUp(self):
        super(DbTestCase, self).setUp()
        # NOTE(mpeterson): Due to how the pecan lib does introspection
        # to find a non-decorated function, it needs a function that will
        # be found in the end. The line below workarounds this limitation.
        self._mock_function = mock.MagicMock()

    def _update_row(self, row):
        self.db_context.session.merge(row)
        self.db_context.session.flush()

    def _test_validate_updates(self, first_entry, second_entry, expected_deps,
                               state=None):
        db.create_pending_row(self.db_context, *first_entry)
        if state:
            row = db.get_all_db_rows(self.db_context)[0]
            row.state = state
            self._update_row(row)

        deps = db.get_pending_or_processing_ops(
            self.db_context, second_entry[1], second_entry[2])
        self.assertEqual(expected_deps, len(deps) != 0)

    def _test_retry_count(self, retry_num, max_retry,
                          expected_retry_count, expected_state):
        # add new pending row
        db.create_pending_row(self.db_context, *self.UPDATE_ROW)

        # update the row with the requested retry_num
        row = db.get_all_db_rows(self.db_context)[0]
        row.retry_count = retry_num - 1
        db.update_pending_db_row_retry(self.db_context, row, max_retry)

        # validate the state and the retry_count of the row
        row = db.get_all_db_rows(self.db_context)[0]
        self.assertEqual(expected_state, row.state)
        self.assertEqual(expected_retry_count, row.retry_count)

    def _test_retry_wrapper(self, decorated_function):
        # NOTE(mpeterson): we want to make sure that it's configured
        # to MAX_RETRIES.
        self.assertEqual(db_api._retry_db_errors.max_retries,
                         db_api.MAX_RETRIES)

        self._test_retry_exceptions(decorated_function,
                                    self._mock_function, False)

    # NOTE(mpeterson): The following function serves to workaround a
    # limitation in the discovery mechanism of pecan lib that does not allow
    # us to create a generic function that decorates on the fly. It needs to
    # be decorated through the decorator directive and not via function
    # composition
    @db_api.retry_if_session_inactive()
    def _decorated_retry_if_session_inactive(self, context):
        self._mock_function()

    def test_retry_if_session_inactive(self):
        self._test_retry_wrapper(self._decorated_retry_if_session_inactive)

    @in_session
    def _test_update_row_state(self, from_state, to_state, dry_flush=False):
        # add new pending row
        db.create_pending_row(self.db_context, *self.UPDATE_ROW)

        mock_flush = mock.MagicMock(
            side_effect=self.db_context.session.flush)

        if dry_flush:
            patch_flush = mock.patch.object(self.db_context.session,
                                            'flush',
                                            side_effect=mock_flush)

        row = db.get_all_db_rows(self.db_context)[0]
        for state in [from_state, to_state]:
            if dry_flush:
                patch_flush.start()

            try:
                # update the row state
                db.update_db_row_state(self.db_context, row, state,
                                       flush=not dry_flush)
            finally:
                if dry_flush:
                    patch_flush.stop()

            # validate the new state
            row = db.get_all_db_rows(self.db_context)[0]
            self.assertEqual(state, row.state)

        return mock_flush

    def test_updates_same_object_uuid(self):
        self._test_validate_updates(self.UPDATE_ROW, self.UPDATE_ROW, True)

    def test_validate_updates_different_object_uuid(self):
        other_row = list(self.UPDATE_ROW)
        other_row[1] += 'a'
        self._test_validate_updates(self.UPDATE_ROW, other_row, False)

    def test_validate_updates_different_object_type(self):
        other_row = list(self.UPDATE_ROW)
        other_row[0] = odl_const.ODL_PORT
        other_row[1] += 'a'
        self._test_validate_updates(self.UPDATE_ROW, other_row, False)

    def test_check_for_older_ops_processing(self):
        self._test_validate_updates(self.UPDATE_ROW, self.UPDATE_ROW, True,
                                    state=odl_const.PROCESSING)

    @in_session
    def test_get_oldest_pending_row_none_when_no_rows(self):
        row = db.get_oldest_pending_db_row_with_lock(self.db_context)
        self.assertIsNone(row)

    @in_session
    def _test_get_oldest_pending_row_none(self, state):
        db.create_pending_row(self.db_context, *self.UPDATE_ROW)
        row = db.get_all_db_rows(self.db_context)[0]
        row.state = state
        self._update_row(row)

        row = db.get_oldest_pending_db_row_with_lock(self.db_context)
        self.assertIsNone(row)

    def test_get_oldest_pending_row_none_when_row_processing(self):
        self._test_get_oldest_pending_row_none(odl_const.PROCESSING)

    def test_get_oldest_pending_row_none_when_row_failed(self):
        self._test_get_oldest_pending_row_none(odl_const.FAILED)

    def test_get_oldest_pending_row_none_when_row_completed(self):
        self._test_get_oldest_pending_row_none(odl_const.COMPLETED)

    @in_session
    def test_get_oldest_pending_row(self):
        db.create_pending_row(self.db_context, *self.UPDATE_ROW)
        row = db.get_oldest_pending_db_row_with_lock(self.db_context)
        self.assertIsNotNone(row)
        self.assertEqual(odl_const.PROCESSING, row.state)

    @in_session
    def test_get_oldest_pending_row_order(self):
        db.create_pending_row(self.db_context, *self.UPDATE_ROW)
        older_row = db.get_all_db_rows(self.db_context)[0]
        older_row.last_retried -= timedelta(minutes=1)
        self._update_row(older_row)

        db.create_pending_row(self.db_context, *self.UPDATE_ROW)
        row = db.get_oldest_pending_db_row_with_lock(self.db_context)
        self.assertEqual(older_row, row)

    @in_session
    def _test_get_oldest_pending_row_with_dep(self, dep_state):
        db.create_pending_row(self.db_context, *self.UPDATE_ROW)
        parent_row = db.get_all_db_rows(self.db_context)[0]
        db.update_db_row_state(self.db_context, parent_row, dep_state)
        db.create_pending_row(self.db_context, *self.UPDATE_ROW,
                              depending_on=[parent_row])
        row = db.get_oldest_pending_db_row_with_lock(self.db_context)
        if row is not None:
            self.assertNotEqual(parent_row.seqnum, row.seqnum)

        return row

    @in_session
    def test_get_oldest_pending_row_when_dep_completed(self):
        row = self._test_get_oldest_pending_row_with_dep(odl_const.COMPLETED)
        self.assertEqual(odl_const.PROCESSING, row.state)

    def test_get_oldest_pending_row_when_dep_failed(self):
        row = self._test_get_oldest_pending_row_with_dep(odl_const.FAILED)
        self.assertEqual(odl_const.PROCESSING, row.state)

    @in_session
    def test_get_oldest_pending_row_returns_parent_when_dep_pending(self):
        db.create_pending_row(self.db_context, *self.UPDATE_ROW)
        parent_row = db.get_all_db_rows(self.db_context)[0]
        db.create_pending_row(self.db_context, *self.UPDATE_ROW,
                              depending_on=[parent_row])
        row = db.get_oldest_pending_db_row_with_lock(self.db_context)
        self.assertEqual(parent_row, row)

    @in_session
    def test_get_oldest_pending_row_none_when_dep_processing(self):
        row = self._test_get_oldest_pending_row_with_dep(odl_const.PROCESSING)
        self.assertIsNone(row)

    def test_get_oldest_pending_row_retries_exceptions(self):
        with mock.patch.object(db, 'aliased') as m:
            self._test_retry_exceptions(db.get_oldest_pending_db_row_with_lock,
                                        m)

    @in_session
    def _test_delete_row(self, by_row=False, by_row_id=False, dry_flush=False):
        db.create_pending_row(self.db_context, *self.UPDATE_ROW)
        db.create_pending_row(self.db_context, *self.UPDATE_ROW)

        rows = db.get_all_db_rows(self.db_context)
        self.assertEqual(len(rows), 2)
        row = rows[-1]

        params = {'flush': not dry_flush}
        if by_row:
            params['row'] = row
        elif by_row_id:
            params['row_id'] = row.seqnum

        mock_flush = None
        if dry_flush:
            patch_flush = mock.patch.object(
                self.db_context.session, 'flush',
                side_effect=self.db_context.session.flush
            )
            mock_flush = patch_flush.start()

        try:
            db.delete_row(self.db_context, **params)
        finally:
            if dry_flush:
                patch_flush.stop()
                self.db_context.session.flush()

        rows = db.get_all_db_rows(self.db_context)
        self.assertEqual(len(rows), 1)
        self.assertNotEqual(row.seqnum, rows[0].seqnum)

        return mock_flush

    def test_delete_row_by_row(self):
        self._test_delete_row(by_row=True)

    def test_delete_row_by_row_id(self):
        self._test_delete_row(by_row_id=True)

    def test_delete_row_by_row_without_flushing(self):
        mock_flush = self._test_delete_row(by_row=True, dry_flush=True)
        mock_flush.assert_not_called()

    def test_create_pending_row(self):
        row = db.create_pending_row(self.db_context, *self.UPDATE_ROW)
        self.assertIsNotNone(row)
        rows = db.get_all_db_rows(self.db_context)
        self.assertTrue(row in rows)

    @in_session
    def _test_delete_rows_by_state_and_time(self, last_retried, row_retention,
                                            state, expected_rows,
                                            dry_delete=False):
        db.create_pending_row(self.db_context, *self.UPDATE_ROW)

        # update state and last retried
        row = db.get_all_db_rows(self.db_context)[-1]
        row.state = state
        row.last_retried = row.last_retried - timedelta(seconds=last_retried)
        self._update_row(row)

        if not dry_delete:
            db.delete_rows_by_state_and_time(self.db_context,
                                             odl_const.COMPLETED,
                                             timedelta(seconds=row_retention))

        # validate the number of rows in the journal
        rows = db.get_all_db_rows(self.db_context)
        self.assertEqual(expected_rows, len(rows))

    def test_delete_completed_rows_no_new_rows(self):
        self._test_delete_rows_by_state_and_time(0, 10, odl_const.COMPLETED, 1)

    def test_delete_completed_rows_one_new_row(self):
        self._test_delete_rows_by_state_and_time(6, 5, odl_const.COMPLETED, 0)

    def test_delete_completed_rows_wrong_state(self):
        self._test_delete_rows_by_state_and_time(10, 8, odl_const.PENDING, 1)

    @in_session
    def test_delete_completed_rows_individually(self):
        self._test_delete_rows_by_state_and_time(
            6, 5, odl_const.COMPLETED, 1, True
        )
        patch_delete = mock.patch.object(
            self.db_context.session, 'delete',
            side_effect=self.db_context.session.delete
        )
        mock_delete = patch_delete.start()
        self.addCleanup(patch_delete.stop)
        self._test_delete_rows_by_state_and_time(
            6, 5, odl_const.COMPLETED, 0
        )
        self.assertEqual(mock_delete.call_count, 2)

    @mock.patch.object(db, 'delete_row', side_effect=db.delete_row)
    def test_delete_completed_rows_without_flush(self, mock_delete_row):
        self._test_delete_rows_by_state_and_time(6, 5, odl_const.COMPLETED, 0)
        self.assertEqual({'flush': False}, mock_delete_row.call_args[1])

    @in_session
    def _test_reset_processing_rows(self, last_retried, max_timedelta,
                                    quantity, dry_reset=False):
        db.create_pending_row(self.db_context, *self.UPDATE_ROW)
        expected_state = odl_const.PROCESSING

        row = db.get_all_db_rows(self.db_context)[-1]
        row.state = expected_state
        row.last_retried = row.last_retried - timedelta(seconds=last_retried)
        self._update_row(row)

        if not dry_reset:
            expected_state = odl_const.PENDING
            reset = db.reset_processing_rows(self.db_context, max_timedelta)
            self.assertIsInstance(reset, int)
            self.assertEqual(reset, quantity)

        rows = db.get_all_db_rows_by_state(self.db_context, expected_state)

        self.assertEqual(len(rows), quantity)
        for row in rows:
            self.assertEqual(row.state, expected_state)

    def test_reset_processing_rows(self):
        self._test_reset_processing_rows(6, 5, 1)

    def test_reset_processing_rows_no_new_rows(self):
        self._test_reset_processing_rows(0, 10, 0)

    @mock.patch.object(db, 'update_db_row_state',
                       side_effect=db.update_db_row_state)
    def test_reset_processing_rows_individually(self, mock_update_row):
        self._test_reset_processing_rows(6, 5, 1, True)
        self._test_reset_processing_rows(6, 5, 2)
        self.assertEqual(mock_update_row.call_count, 2)
        self.assertEqual(mock_update_row.call_args[1], {'flush': False})

    def test_valid_retry_count(self):
        self._test_retry_count(1, 1, 1, odl_const.PENDING)

    def test_invalid_retry_count(self):
        self._test_retry_count(2, 1, 1, odl_const.FAILED)

    def test_update_row_state_to_pending(self):
        self._test_update_row_state(odl_const.PROCESSING, odl_const.PENDING)

    def test_update_row_state_to_processing(self):
        self._test_update_row_state(odl_const.PENDING, odl_const.PROCESSING)

    def test_update_row_state_to_failed(self):
        self._test_update_row_state(odl_const.PROCESSING, odl_const.FAILED)

    def test_update_row_state_to_completed(self):
        self._test_update_row_state(odl_const.PROCESSING, odl_const.COMPLETED)

    def test_update_row_state_to_status_without_flush(self):
        mock_flush = self._test_update_row_state(odl_const.PROCESSING,
                                                 odl_const.COMPLETED,
                                                 dry_flush=True)
        # NOTE(mpeterson): call_count=2 because session.merge() calls flush()
        # and we are changing the status twice
        self.assertEqual(mock_flush.call_count, 2)

    @in_session
    def _test_periodic_task_lock_unlock(self, db_func, existing_state,
                                        expected_state, expected_result,
                                        task='test_task'):
        row = models.OpenDaylightPeriodicTask(state=existing_state,
                                              task=task)
        self.db_context.session.add(row)
        self.db_context.session.flush()

        self.assertEqual(expected_result, db_func(self.db_context,
                                                  task))
        row = self.db_context.session.query(
            models.OpenDaylightPeriodicTask).filter_by(task=task).one()

        self.assertEqual(expected_state, row['state'])

    def test_lock_periodic_task(self):
        self._test_periodic_task_lock_unlock(db.lock_periodic_task,
                                             odl_const.PENDING,
                                             odl_const.PROCESSING,
                                             True)

    def test_lock_periodic_task_fails_when_processing(self):
        self._test_periodic_task_lock_unlock(db.lock_periodic_task,
                                             odl_const.PROCESSING,
                                             odl_const.PROCESSING,
                                             False)

    def test_unlock_periodic_task(self):
        self._test_periodic_task_lock_unlock(db.unlock_periodic_task,
                                             odl_const.PROCESSING,
                                             odl_const.PENDING,
                                             True)

    def test_unlock_periodic_task_fails_when_pending(self):
        self._test_periodic_task_lock_unlock(db.unlock_periodic_task,
                                             odl_const.PENDING,
                                             odl_const.PENDING,
                                             False)

    def test_multiple_row_tasks(self):
        self._test_periodic_task_lock_unlock(db.unlock_periodic_task,
                                             odl_const.PENDING,
                                             odl_const.PENDING,
                                             False)

    def _add_tasks(self, tasks):
        row = []
        for count, task in enumerate(tasks):
            row.append(models.OpenDaylightPeriodicTask(state=odl_const.PENDING,
                                                       task=task))
            self.db_context.session.add(row[count])

        self.db_context.session.flush()

        rows = self.db_context.session.query(
            models.OpenDaylightPeriodicTask).all()
        self.assertEqual(len(tasks), len(rows))

    def _perform_ops_on_all_rows(self, tasks, to_lock):
        if to_lock:
            curr_state = odl_const.PENDING
            exp_state = odl_const.PROCESSING
            func = db.lock_periodic_task
        else:
            exp_state = odl_const.PENDING
            curr_state = odl_const.PROCESSING
            func = db.unlock_periodic_task

        processed = []
        for task in tasks:
            row = self.db_context.session.query(
                models.OpenDaylightPeriodicTask).filter_by(task=task).one()

            self.assertEqual(row['state'], curr_state)
            self.assertTrue(func(self.db_context, task))
            rows = self.db_context.session.query(
                models.OpenDaylightPeriodicTask).filter_by().all()

            processed.append(task)

            for row in rows:
                if row['task'] in processed:
                    self.assertEqual(exp_state, row['state'])
                else:
                    self.assertEqual(curr_state, row['state'])

        self.assertFalse(func(self.db_context, tasks[-1]))

    def test_multiple_row_tasks_lock_unlock(self):
        task1 = 'test_random_task'
        task2 = 'random_task_random'
        task3 = 'task_test_random'
        tasks = [task1, task2, task3]
        self._add_tasks(tasks)
        self._perform_ops_on_all_rows(tasks, to_lock=True)
        self._perform_ops_on_all_rows(tasks, to_lock=False)



#
# Copyright (C) 2016 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

import functools

from neutron.tests.unit.extensions import test_l3
from neutron.tests.unit.plugins.ml2 import test_plugin
from neutron_lib import constants as q_const

from networking_odl.common import constants as odl_const
from networking_odl.tests.functional import base


class _TestL3Base(test_l3.L3NatTestCaseMixin, base.OdlTestsBase):

    # Override default behavior so that extension manager is used, otherwise
    # we can't test security groups.
    def setup_parent(self):
        """Perform parent setup with the common plugin configuration class."""
        ext_mgr = test_l3.L3TestExtensionManager()

        # Ensure that the parent setup can be called without arguments
        # by the common configuration setUp.
        parent_setup = functools.partial(
            super(test_plugin.Ml2PluginV2TestCase, self).setUp,
            plugin=test_plugin.PLUGIN_NAME,
            ext_mgr=ext_mgr,
            service_plugins={'l3_plugin_name': self.l3_plugin},

        )
        self.useFixture(test_plugin.Ml2ConfFixture(parent_setup))

    def test_router_create(self):
        with self.router() as router:
            self.assert_resource_created(odl_const.ODL_ROUTER, router)

    def test_router_update(self):
        with self.router() as router:
            self.resource_update_test(odl_const.ODL_ROUTER, router)

    def test_router_delete(self):
        with self.router() as router:
            self.resource_delete_test(odl_const.ODL_ROUTER, router)

    def test_floatingip_create(self):
        with self.floatingip_with_assoc() as fip:
            self.assert_resource_created(odl_const.ODL_FLOATINGIP, fip)

        # Test FIP was deleted since the code creating the FIP deletes it
        # once the context block exists.
        odl_fip = self.get_odl_resource(odl_const.ODL_FLOATINGIP, fip)
        self.assertIsNone(odl_fip)

    def test_floatingip_status_with_port(self):
        with self.floatingip_with_assoc() as fip:
            self.assertEqual(
                q_const.FLOATINGIP_STATUS_ACTIVE,
                fip['floatingip']['status'])

    def test_floatingip_status_without_port(self):
        with self.subnet() as subnet:
            with self.floatingip_no_assoc(subnet) as fip:
                # status should be down when floating ip
                # is not associated to any port
                self.assertEqual(
                    q_const.FLOATINGIP_STATUS_DOWN,
                    fip['floatingip']['status'])

    def test_floatingip_dissociate_port(self):
        with self.floatingip_with_assoc() as fip:
            portid = fip['floatingip']['port_id']
            self.assertIsNotNone(portid)
            self._delete(odl_const.ODL_PORTS, portid)
            updated_fip = self.get_odl_resource(odl_const.ODL_FLOATINGIP, fip)
            self.assertNotIn('port_id', updated_fip['floatingip'].keys())


class TestL3PluginV2(base.V2DriverAdjustment, _TestL3Base,
                     test_plugin.Ml2PluginV2TestCase):
    _mechanism_drivers = ['opendaylight_v2']
    l3_plugin = 'odl-router_v2'



# Copyright (c) 2017 Brocade Communication Systems
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from networking_sfc.extensions import sfc as sfc_const
from networking_sfc.services.sfc.drivers import base as sfc_driver

from networking_odl.common import constants as odl_const
from networking_odl.common import postcommit
from networking_odl.journal import full_sync
from networking_odl.journal import journal

LOG = logging.getLogger(__name__)

SFC_RESOURCES = {
    odl_const.ODL_SFC_PORT_PAIR: odl_const.NETWORKING_SFC_PORT_PAIRS,
    odl_const.ODL_SFC_PORT_PAIR_GROUP:
        odl_const.NETWORKING_SFC_PORT_PAIR_GROUPS,
    odl_const.ODL_SFC_PORT_CHAIN: odl_const.NETWORKING_SFC_PORT_CHAINS
}


@postcommit.add_postcommit('port_pair', 'port_pair_group', 'port_chain')
class OpenDaylightSFCDriverV2(sfc_driver.SfcDriverBase):
    """OpenDaylight SFC Driver (Version 2) for networking-sfc.

    Driver sends REST request for Networking SFC Resources (Port Pair,
    Port Pair Group & Port Chain) to OpenDaylight Neutron Northbound.
    OpenDaylight Neutron Northbound has API's defined for these resources
    based on the Networking SFC APIs.
    """

    def initialize(self):
        LOG.debug("Initializing OpenDaylight Networking SFC driver(Version 2)")
        self.journal = journal.OpenDaylightJournalThread()
        full_sync.register(sfc_const.SFC_EXT, SFC_RESOURCES)

    @staticmethod
    def _record_in_journal(context, object_type, operation, data=None):
        if data is None:
            data = context.current
        journal.record(context._plugin_context, object_type,
                       context.current['id'], operation, data)

    @log_helpers.log_method_call
    def create_port_pair_precommit(self, context):
        OpenDaylightSFCDriverV2._record_in_journal(
            context, odl_const.ODL_SFC_PORT_PAIR, odl_const.ODL_CREATE)

    @log_helpers.log_method_call
    def create_port_pair_group_precommit(self, context):
        OpenDaylightSFCDriverV2._record_in_journal(
            context, odl_const.ODL_SFC_PORT_PAIR_GROUP, odl_const.ODL_CREATE)

    @log_helpers.log_method_call
    def create_port_chain_precommit(self, context):
        OpenDaylightSFCDriverV2._record_in_journal(
            context, odl_const.ODL_SFC_PORT_CHAIN, odl_const.ODL_CREATE)

    @log_helpers.log_method_call
    def update_port_pair_precommit(self, context):
        OpenDaylightSFCDriverV2._record_in_journal(
            context, odl_const.ODL_SFC_PORT_PAIR, odl_const.ODL_UPDATE)

    @log_helpers.log_method_call
    def update_port_pair_group_precommit(self, context):
        OpenDaylightSFCDriverV2._record_in_journal(
            context, odl_const.ODL_SFC_PORT_PAIR_GROUP, odl_const.ODL_UPDATE)

    @log_helpers.log_method_call
    def update_port_chain_precommit(self, context):
        OpenDaylightSFCDriverV2._record_in_journal(
            context, odl_const.ODL_SFC_PORT_CHAIN, odl_const.ODL_UPDATE)

    @log_helpers.log_method_call
    def delete_port_pair_precommit(self, context):
        OpenDaylightSFCDriverV2._record_in_journal(
            context, odl_const.ODL_SFC_PORT_PAIR, odl_const.ODL_DELETE,
            data=[])

    @log_helpers.log_method_call
    def delete_port_pair_group_precommit(self, context):
        OpenDaylightSFCDriverV2._record_in_journal(
            context, odl_const.ODL_SFC_PORT_PAIR_GROUP, odl_const.ODL_DELETE,
            data=[])

    @log_helpers.log_method_call
    def delete_port_chain_precommit(self, context):
        OpenDaylightSFCDriverV2._record_in_journal(
            context, odl_const.ODL_SFC_PORT_CHAIN, odl_const.ODL_DELETE,
            data=[])

    # Need to implement these methods, else driver loading fails with error
    # complaining about no abstract method implementation present.
    @log_helpers.log_method_call
    def create_port_pair(self, context):
        super(OpenDaylightSFCDriverV2, self).create_port_pair(context)

    @log_helpers.log_method_call
    def create_port_pair_group(self, context):
        super(OpenDaylightSFCDriverV2, self).create_port_pair_group(context)

    @log_helpers.log_method_call
    def create_port_chain(self, context):
        super(OpenDaylightSFCDriverV2, self).create_port_chain(context)

    @log_helpers.log_method_call
    def update_port_pair(self, context):
        super(OpenDaylightSFCDriverV2, self).update_port_pair(context)

    @log_helpers.log_method_call
    def update_port_pair_group(self, context):
        super(OpenDaylightSFCDriverV2, self).update_port_pair_group(context)

    @log_helpers.log_method_call
    def update_port_chain(self, context):
        super(OpenDaylightSFCDriverV2, self).update_port_chain(context)

    @log_helpers.log_method_call
    def delete_port_pair(self, context):
        super(OpenDaylightSFCDriverV2, self).delete_port_pair(context)

    @log_helpers.log_method_call
    def delete_port_pair_group(self, context):
        super(OpenDaylightSFCDriverV2, self).delete_port_pair_group(context)

    @log_helpers.log_method_call
    def delete_port_chain(self, context):
        super(OpenDaylightSFCDriverV2, self).delete_port_chain(context)



#
# Copyright (C) 2016 Intel Corp. Isaku Yamahata <isaku.yamahata@gmail com>
# All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

import testscenarios

from networking_odl.common import constants as const
from networking_odl.db import db
from networking_odl.journal import dependency_validations
from networking_odl.tests.unit import test_base_db


load_tests = testscenarios.load_tests_apply_scenarios


_NET_ID = 'NET_ID'
_NET_DATA = {'id': _NET_ID}
_SUBNET_ID = 'SUBNET_ID'
_SUBNET_DATA = {'network_id': _NET_ID}
_PORT_ID = 'PORT_ID'
_PORT_DATA = {'network_id': _NET_ID,
              'fixed_ips': [{'subnet_id': _SUBNET_ID}]}
_PORT_DATA_DUPLICATE_SUBNET = {
    'network_id': _NET_ID,
    'fixed_ips': [{'subnet_id': _SUBNET_ID},
                  {'subnet_id': _SUBNET_ID}]
}
_ROUTER_ID = 'ROUTER_ID'
_ROUTER_DATA = {'id': 'ROUTER_ID',
                'gw_port_id': 'GW_PORT_ID'}
_L2GW_ID = 'l2gw_id'
_L2GW_DATA = {'id': _L2GW_ID}
_L2GWCONN_ID = 'l2gwconn_id'
_L2GWCONN_DATA = {'id': _L2GWCONN_ID,
                  'network_id': _NET_ID,
                  'gateway_id': _L2GW_ID}
_TRUNK_ID = 'TRUNK_ID'
_SUBPORT_ID = 'CPORT_ID'
_TRUNK_DATA = {'trunk_id': _TRUNK_ID,
               'port_id': _PORT_ID,
               'sub_ports': [{'port_id': _SUBPORT_ID}]}
_BGPVPN_ID = 'BGPVPN_ID'
_SG_ID = 'SG_ID'
_SG_DATA = {'id': _SG_ID}
_SG_RULE_ID = 'SG_RULE_ID'
_SG_RULE_DATA = {'id': _SG_RULE_ID,
                 'security_group_id': _SG_ID}


def get_data(res_type, operation):
    if res_type == const.ODL_NETWORK:
        return [_NET_DATA]
    elif res_type == const.ODL_SUBNET:
        if operation == const.ODL_DELETE:
            return [[_NET_ID]]
        return [_SUBNET_DATA]
    elif res_type == const.ODL_PORT:
        # TODO(yamahata): test case of (ODL_port, ODL_DELETE) is missing
        if operation == const.ODL_DELETE:
            return [[_NET_ID, _SUBNET_ID]]
        return [_PORT_DATA, _PORT_DATA_DUPLICATE_SUBNET]
    elif res_type == const.ODL_ROUTER:
        return [_ROUTER_DATA]
    elif res_type == const.ODL_L2GATEWAY:
        return [_L2GW_DATA]
    elif res_type == const.ODL_L2GATEWAY_CONNECTION:
        return [_L2GWCONN_DATA]
    elif res_type == const.ODL_TRUNK:
        if operation == const.ODL_DELETE:
            return [[_PORT_ID, _SUBPORT_ID]]
        return [_TRUNK_DATA]
    elif res_type == const.ODL_BGPVPN:
        if operation == const.ODL_DELETE:
            return [[_NET_ID, _ROUTER_ID]]
        else:
            routers = []
            networks = []
            if operation == const.ODL_UPDATE:
                routers = [_ROUTER_ID]
                networks = [_NET_ID]
            return [{'id': _BGPVPN_ID, 'networks': networks,
                     'routers': routers,
                     'route_distinguishers': ['100:1']}]
    elif res_type == const.ODL_SG:
        return [_SG_DATA]
    elif res_type == const.ODL_SG_RULE:
        if operation == const.ODL_DELETE:
            return [[_SG_RULE_ID]]
        return [_SG_RULE_DATA]
    return [[]]


def subnet_fail_network_dep(net_op, subnet_op):
    return {'expected': 1,
            'first_type': const.ODL_NETWORK,
            'first_operation': net_op,
            'first_id': _NET_ID,
            'second_type': const.ODL_SUBNET,
            'second_operation': subnet_op,
            'second_id': _SUBNET_ID}


def subnet_succeed_network_dep(net_op, subnet_op):
    return {'expected': 0,
            'first_type': const.ODL_SUBNET,
            'first_operation': subnet_op,
            'first_id': _SUBNET_ID,
            'second_type': const.ODL_NETWORK,
            'second_operation': net_op,
            'second_id': _NET_ID}


# TODO(vthapar) add tests for l2gw dependency validations
class BaseDependencyValidationsTestCase(object):
    def test_dependency(self):
        db.create_pending_row(
            self.db_context, self.first_type, self.first_id,
            self.first_operation,
            get_data(self.first_type, self.first_operation))
        for data in get_data(self.second_type, self.second_operation):
            deps = dependency_validations.calculate(
                self.db_context, self.second_operation, self.second_type,
                self.second_id, data)
            self.assertEqual(self.expected, len(deps))


class SubnetDependencyValidationsTestCase(
        test_base_db.ODLBaseDbTestCase, BaseDependencyValidationsTestCase):
    scenarios = (
        ("subnet_create_depends_on_older_network_create",
         subnet_fail_network_dep(const.ODL_CREATE, const.ODL_CREATE)),
        ("subnet_create_depends_on_older_network_update",
         subnet_fail_network_dep(const.ODL_UPDATE, const.ODL_CREATE)),
        ("subnet_create_depends_on_older_network_delete",
         subnet_fail_network_dep(const.ODL_DELETE, const.ODL_CREATE)),
        ("subnet_create_doesnt_depend_on_newer_network_create",
         subnet_succeed_network_dep(const.ODL_CREATE, const.ODL_CREATE)),
        ("subnet_create_doesnt_depend_on_newer_network_update",
         subnet_succeed_network_dep(const.ODL_UPDATE, const.ODL_CREATE)),
        ("subnet_create_doesnt_depend_on_newer_network_delete",
         subnet_succeed_network_dep(const.ODL_DELETE, const.ODL_CREATE)),
        ("subnet_update_depends_on_older_network_create",
         subnet_fail_network_dep(const.ODL_CREATE, const.ODL_UPDATE)),
        ("subnet_update_depends_on_older_network_update",
         subnet_fail_network_dep(const.ODL_UPDATE, const.ODL_UPDATE)),
        ("subnet_update_depends_on_older_network_delete",
         subnet_fail_network_dep(const.ODL_DELETE, const.ODL_UPDATE)),
        ("subnet_update_doesnt_depend_on_newer_network_create",
         subnet_succeed_network_dep(const.ODL_CREATE, const.ODL_UPDATE)),
        ("subnet_update_doesnt_depend_on_newer_network_update",
         subnet_succeed_network_dep(const.ODL_UPDATE, const.ODL_UPDATE)),
        ("subnet_update_doesnt_depend_on_newer_network_delete",
         subnet_succeed_network_dep(const.ODL_DELETE, const.ODL_UPDATE)),
        ("subnet_delete_doesnt_depend_on_older_network_create",
         subnet_succeed_network_dep(const.ODL_CREATE, const.ODL_DELETE)),
        ("subnet_delete_doesnt_depend_on_older_network_update",
         subnet_succeed_network_dep(const.ODL_UPDATE, const.ODL_DELETE)),
        ("subnet_delete_doesnt_depend_on_newer_network_create",
         subnet_succeed_network_dep(const.ODL_CREATE, const.ODL_DELETE)),
        ("subnet_delete_doesnt_depend_on_newer_network_update",
         subnet_succeed_network_dep(const.ODL_UPDATE, const.ODL_DELETE)),
    )


def security_rule_fail_security_group_dep(sg_op, sgr_op):
    return {'expected': 1,
            'first_type': const.ODL_SG,
            'first_operation': sg_op,
            'first_id': _SG_ID,
            'second_type': const.ODL_SG_RULE,
            'second_operation': sgr_op,
            'second_id': _SG_RULE_ID}


def security_rule_succeed_security_group_dep(sg_op, sgr_op):
    return {'expected': 0,
            'first_type': const.ODL_SG_RULE,
            'first_operation': sgr_op,
            'first_id': _SG_RULE_ID,
            'second_type': const.ODL_SG,
            'second_operation': sg_op,
            'second_id': _SG_ID}


class SecurityRuleDependencyValidationsTestCase(
        test_base_db.ODLBaseDbTestCase, BaseDependencyValidationsTestCase):
    scenarios = (
        ("security_rule_create_depends_on_older_security_group_create",
         security_rule_fail_security_group_dep(const.ODL_CREATE,
                                               const.ODL_CREATE)),
        ("security_rule_create_depends_on_older_security_group_update",
         security_rule_fail_security_group_dep(const.ODL_UPDATE,
                                               const.ODL_CREATE)),
        ("security_rule_create_depends_on_older_security_group_delete",
         security_rule_fail_security_group_dep(const.ODL_DELETE,
                                               const.ODL_CREATE)),
        ("security_rule_create_doesnt_depend_on_newer_security_group_create",
         security_rule_succeed_security_group_dep(const.ODL_CREATE,
                                                  const.ODL_CREATE)),
        ("security_rule_create_doesnt_depend_on_newer_security_group_update",
         security_rule_succeed_security_group_dep(const.ODL_UPDATE,
                                                  const.ODL_CREATE)),
        ("security_rule_create_doesnt_depend_on_newer_security_group_delete",
         security_rule_succeed_security_group_dep(const.ODL_DELETE,
                                                  const.ODL_CREATE)),
        ("security_rule_update_depends_on_older_security_group_create",
         security_rule_fail_security_group_dep(const.ODL_CREATE,
                                               const.ODL_UPDATE)),
        ("security_rule_update_depends_on_older_security_group_update",
         security_rule_fail_security_group_dep(const.ODL_UPDATE,
                                               const.ODL_UPDATE)),
        ("security_rule_update_depends_on_older_security_group_delete",
         security_rule_fail_security_group_dep(const.ODL_DELETE,
                                               const.ODL_UPDATE)),
        ("security_rule_update_doesnt_depend_on_newer_security_group_create",
         security_rule_succeed_security_group_dep(const.ODL_CREATE,
                                                  const.ODL_UPDATE)),
        ("security_rule_update_doesnt_depend_on_newer_security_group_update",
         security_rule_succeed_security_group_dep(const.ODL_UPDATE,
                                                  const.ODL_UPDATE)),
        ("security_rule_update_doesnt_depend_on_newer_security_group_delete",
         security_rule_succeed_security_group_dep(const.ODL_DELETE,
                                                  const.ODL_UPDATE)),
        ("security_rule_delete_doesnt_depend_on_older_security_group_create",
         security_rule_succeed_security_group_dep(const.ODL_CREATE,
                                                  const.ODL_DELETE)),
        ("security_rule_delete_doesnt_depend_on_older_security_group_update",
         security_rule_succeed_security_group_dep(const.ODL_UPDATE,
                                                  const.ODL_DELETE)),
        ("security_rule_delete_doesnt_depend_on_newer_security_group_create",
         security_rule_succeed_security_group_dep(const.ODL_CREATE,
                                                  const.ODL_DELETE)),
        ("security_rule_delete_doesnt_depend_on_newer_security_group_update",
         security_rule_succeed_security_group_dep(const.ODL_UPDATE,
                                                  const.ODL_DELETE)),
    )


def port_fail_network_dep(net_op, port_op):
    return {'expected': 1,
            'first_type': const.ODL_NETWORK,
            'first_operation': net_op,
            'first_id': _NET_ID,
            'second_type': const.ODL_PORT,
            'second_operation': port_op,
            'second_id': _PORT_ID}


def port_succeed_network_dep(net_op, port_op):
    return {'expected': 0,
            'first_type': const.ODL_PORT,
            'first_operation': port_op,
            'first_id': _PORT_ID,
            'second_type': const.ODL_NETWORK,
            'second_operation': net_op,
            'second_id': _NET_ID}


def port_fail_subnet_dep(subnet_op, port_op):
    return {'expected': 1,
            'first_type': const.ODL_SUBNET,
            'first_operation': subnet_op,
            'first_id': _SUBNET_ID,
            'second_type': const.ODL_PORT,
            'second_operation': port_op,
            'second_id': _PORT_ID}


def port_succeed_subnet_dep(subnet_op, port_op):
    return {'expected': 0,
            'first_type': const.ODL_PORT,
            'first_operation': port_op,
            'first_id': _PORT_ID,
            'second_type': const.ODL_SUBNET,
            'second_operation': subnet_op,
            'second_id': _SUBNET_ID}


class PortDependencyValidationsTestCase(
        test_base_db.ODLBaseDbTestCase, BaseDependencyValidationsTestCase):
    scenarios = (
        ("port_create_depends_on_older_network_create",
         port_fail_network_dep(const.ODL_CREATE, const.ODL_CREATE)),
        ("port_create_depends_on_older_network_update",
         port_fail_network_dep(const.ODL_UPDATE, const.ODL_CREATE)),
        ("port_create_depends_on_older_network_delete",
         port_fail_network_dep(const.ODL_DELETE, const.ODL_CREATE)),
        ("port_create_doesnt_depend_on_newer_network_create",
         port_succeed_network_dep(const.ODL_CREATE, const.ODL_CREATE)),
        ("port_create_doesnt_depend_on_newer_network_update",
         port_succeed_network_dep(const.ODL_UPDATE, const.ODL_CREATE)),
        ("port_create_doesnt_depend_on_newer_network_delete",
         port_succeed_network_dep(const.ODL_DELETE, const.ODL_CREATE)),
        ("port_update_depends_on_older_network_create",
         port_fail_network_dep(const.ODL_CREATE, const.ODL_UPDATE)),
        ("port_update_depends_on_older_network_update",
         port_fail_network_dep(const.ODL_UPDATE, const.ODL_UPDATE)),
        ("port_update_depends_on_older_network_delete",
         port_fail_network_dep(const.ODL_DELETE, const.ODL_UPDATE)),
        ("port_update_doesnt_depend_on_newer_network_create",
         port_succeed_network_dep(const.ODL_CREATE, const.ODL_UPDATE)),
        ("port_update_doesnt_depend_on_newer_network_update",
         port_succeed_network_dep(const.ODL_UPDATE, const.ODL_UPDATE)),
        ("port_update_doesnt_depend_on_newer_network_delete",
         port_succeed_network_dep(const.ODL_DELETE, const.ODL_UPDATE)),
        ("port_create_depends_on_older_subnet_create",
         port_fail_subnet_dep(const.ODL_CREATE, const.ODL_CREATE)),
        ("port_create_depends_on_older_subnet_update",
         port_fail_subnet_dep(const.ODL_UPDATE, const.ODL_CREATE)),
        ("port_create_depends_on_older_subnet_delete",
         port_fail_subnet_dep(const.ODL_DELETE, const.ODL_CREATE)),
        ("port_create_doesnt_depend_on_newer_subnet_create",
         port_succeed_subnet_dep(const.ODL_CREATE, const.ODL_CREATE)),
        ("port_create_doesnt_depend_on_newer_subnet_update",
         port_succeed_subnet_dep(const.ODL_UPDATE, const.ODL_CREATE)),
        ("port_create_doesnt_depend_on_newer_subnet_delete",
         port_succeed_subnet_dep(const.ODL_DELETE, const.ODL_CREATE)),
        ("port_update_depends_on_older_subnet_create",
         port_fail_subnet_dep(const.ODL_CREATE, const.ODL_UPDATE)),
        ("port_update_depends_on_older_subnet_update",
         port_fail_subnet_dep(const.ODL_UPDATE, const.ODL_UPDATE)),
        ("port_update_depends_on_older_subnet_delete",
         port_fail_subnet_dep(const.ODL_DELETE, const.ODL_UPDATE)),
        ("port_update_doesnt_depend_on_newer_subnet_create",
         port_succeed_subnet_dep(const.ODL_CREATE, const.ODL_UPDATE)),
        ("port_update_doesnt_depend_on_newer_subnet_update",
         port_succeed_subnet_dep(const.ODL_UPDATE, const.ODL_UPDATE)),
        ("port_update_doesnt_depend_on_newer_subnet_delete",
         port_succeed_subnet_dep(const.ODL_DELETE, const.ODL_UPDATE)),
    )


def trunk_dep(first_type, second_type, first_op, second_op, result,
              sub_port=False):
    expected = {'fail': 1, 'pass': 0}
    port_id = _SUBPORT_ID if sub_port else _PORT_ID
    type_id = {const.ODL_PORT: port_id,
               const.ODL_TRUNK: _TRUNK_ID}
    return {'expected': expected[result],
            'first_type': first_type,
            'first_operation': first_op,
            'first_id': type_id[first_type],
            'second_type': second_type,
            'second_operation': second_op,
            'second_id': type_id[second_type]}


class TrunkDependencyValidationsTestCase(
        test_base_db.ODLBaseDbTestCase, BaseDependencyValidationsTestCase):
    scenarios = (
        ("trunk_create_depends_on_older_port_create",
         trunk_dep(const.ODL_PORT, const.ODL_TRUNK,
                   const.ODL_CREATE, const.ODL_CREATE, 'fail')),
        ("trunk_create_doesnt_depend_on_newer_port_create",
         trunk_dep(const.ODL_TRUNK, const.ODL_PORT,
                   const.ODL_CREATE, const.ODL_CREATE, 'pass')),
        ("trunk_create_doesnt_depend_on_port_update",
         trunk_dep(const.ODL_TRUNK, const.ODL_PORT,
                   const.ODL_CREATE, const.ODL_UPDATE, 'pass')),
        ("trunk_create_doesnt_depend_on_newer_port_delete",
         trunk_dep(const.ODL_TRUNK, const.ODL_PORT,
                   const.ODL_CREATE, const.ODL_DELETE, 'pass')),
        # TODO(vthapar): add more/better validations for subport
        # trunk update means subport add/delete
        ("trunk_update_depends_on_older_trunk_create",
         trunk_dep(const.ODL_TRUNK, const.ODL_TRUNK,
                   const.ODL_CREATE, const.ODL_UPDATE, 'fail', True)),
        ("trunk_update_depends_on_older_port_create",
         trunk_dep(const.ODL_PORT, const.ODL_TRUNK,
                   const.ODL_CREATE, const.ODL_UPDATE, 'fail', True)),
        ("trunk_update_doesnt_depend_on_newer_port_create",
         trunk_dep(const.ODL_TRUNK, const.ODL_PORT,
                   const.ODL_UPDATE, const.ODL_CREATE, 'pass', True)),
        ("trunk_update_doesnt_depend_on_port_update",
         trunk_dep(const.ODL_TRUNK, const.ODL_PORT,
                   const.ODL_UPDATE, const.ODL_UPDATE, 'pass', True)),
        ("trunk_update_doesnt_depend_on_newer_port_delete",
         trunk_dep(const.ODL_TRUNK, const.ODL_PORT,
                   const.ODL_UPDATE, const.ODL_DELETE, 'pass', True)),
        # trunk delete cases
        ("trunk_delete_depends_on_older_trunk_create",
         trunk_dep(const.ODL_TRUNK, const.ODL_TRUNK,
                   const.ODL_CREATE, const.ODL_DELETE, 'fail', True)),
        ("trunk_delete_depends_on_older_trunk_update",
         trunk_dep(const.ODL_TRUNK, const.ODL_TRUNK,
                   const.ODL_UPDATE, const.ODL_DELETE, 'fail', True)),
        ("trunk_delete_doesnt_depend_on_older_port_create",
         trunk_dep(const.ODL_PORT, const.ODL_TRUNK,
                   const.ODL_CREATE, const.ODL_DELETE, 'pass')),
    )


def l2gw_dep(first_type, second_type, first_op, second_op, result):
    expected = {'fail': 1, 'pass': 0}
    type_id = {const.ODL_NETWORK: _NET_ID,
               const.ODL_L2GATEWAY: _L2GW_ID,
               const.ODL_L2GATEWAY_CONNECTION: _L2GWCONN_ID}
    return {'expected': expected[result],
            'first_type': first_type,
            'first_operation': first_op,
            'first_id': type_id[first_type],
            'second_type': second_type,
            'second_operation': second_op,
            'second_id': type_id[second_type]}


class L2GWDependencyValidationsTestCase(
        test_base_db.ODLBaseDbTestCase, BaseDependencyValidationsTestCase):
    scenarios = (
        ("L2GWConn_create_depends_on_older_network_create",
         l2gw_dep(const.ODL_NETWORK, const.ODL_L2GATEWAY_CONNECTION,
                  const.ODL_CREATE, const.ODL_CREATE, 'fail')),
        ("L2GWConn_create_depends_on_older_L2GW_create",
         l2gw_dep(const.ODL_L2GATEWAY, const.ODL_L2GATEWAY_CONNECTION,
                  const.ODL_CREATE, const.ODL_CREATE, 'fail')),
        ("L2GWConn_create_doesnt_depend_on_newer_network_create",
         l2gw_dep(const.ODL_L2GATEWAY_CONNECTION, const.ODL_NETWORK,
                  const.ODL_CREATE, const.ODL_CREATE, 'pass')),
        ("L2GWConn_create_doesnt_depend_on_newer_L2GW_create",
         l2gw_dep(const.ODL_L2GATEWAY_CONNECTION, const.ODL_L2GATEWAY,
                  const.ODL_CREATE, const.ODL_CREATE, 'pass')),
    )


# TODO(vthapar): Refactor *_dep into a common method
def bgpvpn_dep(first_type, second_type, first_op, second_op, result):
    expected = {'fail': 1, 'pass': 0}
    type_id = {const.ODL_NETWORK: _NET_ID,
               const.ODL_ROUTER: _ROUTER_ID,
               const.ODL_BGPVPN: _BGPVPN_ID}
    return {'expected': expected[result],
            'first_type': first_type,
            'first_operation': first_op,
            'first_id': type_id[first_type],
            'second_type': second_type,
            'second_operation': second_op,
            'second_id': type_id[second_type]}


class BGPVPNDependencyValidationsTestCase(
        test_base_db.ODLBaseDbTestCase, BaseDependencyValidationsTestCase):
    scenarios = (
        ("bgpvpn_create_doesnt_depend_on_older_network_create",
         bgpvpn_dep(const.ODL_NETWORK, const.ODL_BGPVPN,
                    const.ODL_CREATE, const.ODL_CREATE, 'pass')),
        ("bgpvpn_create_doesnt_depend_on_newer_network_create",
         bgpvpn_dep(const.ODL_BGPVPN, const.ODL_NETWORK,
                    const.ODL_CREATE, const.ODL_CREATE, 'pass')),
        ("bgpvpn_create_doesnt_depend_on_older_router_create",
         bgpvpn_dep(const.ODL_ROUTER, const.ODL_BGPVPN,
                    const.ODL_CREATE, const.ODL_CREATE, 'pass')),
        ("bgpvpn_create_doesnt_depend_on_newer_router_create",
         bgpvpn_dep(const.ODL_BGPVPN, const.ODL_ROUTER,
                    const.ODL_CREATE, const.ODL_CREATE, 'pass')),
        ("bgpvpn_update_depends_on_older_bgpvpn_create",
         bgpvpn_dep(const.ODL_BGPVPN, const.ODL_BGPVPN,
                    const.ODL_CREATE, const.ODL_UPDATE, 'fail')),
        ("bgpvpn_update_depends_on_older_network_create",
         bgpvpn_dep(const.ODL_NETWORK, const.ODL_BGPVPN,
                    const.ODL_CREATE, const.ODL_UPDATE, 'fail')),
        ("bgpvpn_update_doesnt_depend_on_newer_network_create",
         bgpvpn_dep(const.ODL_BGPVPN, const.ODL_NETWORK,
                    const.ODL_UPDATE, const.ODL_CREATE, 'pass')),
        ("bgpvpn_update_depends_on_older_router_create",
         bgpvpn_dep(const.ODL_ROUTER, const.ODL_BGPVPN,
                    const.ODL_CREATE, const.ODL_UPDATE, 'fail')),
        ("bgpvpn_update_doesnt_depend_on_newer_router_create",
         bgpvpn_dep(const.ODL_BGPVPN, const.ODL_ROUTER,
                    const.ODL_UPDATE, const.ODL_CREATE, 'pass')),
        # bgpvpn delete cases
        ("bgpvpn_delete_depends_on_older_bgpvpn_create",
         bgpvpn_dep(const.ODL_BGPVPN, const.ODL_BGPVPN,
                    const.ODL_CREATE, const.ODL_DELETE, 'fail')),
        ("bgpvpn_delete_depends_on_older_bgpvpn_update",
         bgpvpn_dep(const.ODL_BGPVPN, const.ODL_BGPVPN,
                    const.ODL_UPDATE, const.ODL_DELETE, 'fail')),
        ("bgpvpn_delete_doesnt_depend_on_older_network_create",
         bgpvpn_dep(const.ODL_NETWORK, const.ODL_BGPVPN,
                    const.ODL_CREATE, const.ODL_DELETE, 'pass')),
        ("bgpvpn_delete_doesnt_depend_on_older_router_create",
         bgpvpn_dep(const.ODL_ROUTER, const.ODL_BGPVPN,
                    const.ODL_CREATE, const.ODL_DELETE, 'pass')),
    )



# Copyright (c) 2015-2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from unittest import mock

import fixtures

from neutron.tests import base
from neutron_lib.callbacks import registry
from neutron_lib import fixture as nl_fixture
from oslo_config import cfg
from oslo_config import fixture as config_fixture

from networking_odl.common import odl_features
from networking_odl.journal import full_sync
from networking_odl.journal import journal
from networking_odl.journal import periodic_task
from networking_odl.ml2 import pseudo_agentdb_binding


class DietTestCase(base.DietTestCase):

    def patch(self, target, name, *args, **kwargs):
        context = mock.patch.object(target, name, *args, **kwargs)
        patch = context.start()
        self.addCleanup(context.stop)
        return patch


class OpenDaylightRestClientFixture(fixtures.Fixture):
    # Set URL/user/pass so init doesn't throw a cfg required error.
    # They are not used in these tests since requests.request is overwritten.
    def _setUp(self):
        super(OpenDaylightRestClientFixture, self)._setUp()
        self.cfg = self.useFixture(config_fixture.Config())
        mock.patch('requests.sessions.Session.request').start()
        self.cfg.config(url='http://localhost:8080/controller/nb/v2/neutron',
                        group='ml2_odl')
        self.cfg.config(username='someuser', group='ml2_odl')
        self.cfg.config(password='somepass', group='ml2_odl')
        self.cfg.config(port_binding_controller='legacy-port-binding',
                        group='ml2_odl')


class OpenDaylightRestClientGlobalFixture(fixtures.Fixture):
    def __init__(self, global_client):
        super(OpenDaylightRestClientGlobalFixture, self).__init__()
        self._global_client = global_client

    def _setUp(self):
        super(OpenDaylightRestClientGlobalFixture, self)._setUp()
        mock.patch.object(self._global_client, 'get_client').start()


class OpenDaylightFeaturesFixture(fixtures.Fixture):
    def _setUp(self):
        super(OpenDaylightFeaturesFixture, self)._setUp()
        self.cfg = self.useFixture(config_fixture.Config())
        if cfg.CONF.ml2_odl.url is None:
            self.cfg.config(url='http://127.0.0.1:9999', group='ml2_odl')
        if cfg.CONF.ml2_odl.username is None:
            self.cfg.config(username='someuser', group='ml2_odl')
        if cfg.CONF.ml2_odl.password is None:
            self.cfg.config(password='somepass', group='ml2_odl')
        # make sure _fetch_features is not called, it'll block the main thread
        self.cfg.config(odl_features_json='{"features": {"feature": []}}',
                        group='ml2_odl')
        odl_features.init()
        self.addCleanup(odl_features.deinit)


class OpenDaylightJournalThreadFixture(fixtures.Fixture):
    def _setUp(self):
        super(OpenDaylightJournalThreadFixture, self)._setUp()
        self.journal_thread_mock = mock.patch.object(
            journal.OpenDaylightJournalThread, 'start')
        self.journal_thread_mock.start()
        self.pidfile_fixture = self.useFixture(JournalWorkerPidFileFixture())

    def remock_atexit(self):
        self.pidfile_fixture.atexit_mock.stop()
        return self.pidfile_fixture.atexit_mock.start()


class JournalWorkerPidFileFixture(fixtures.Fixture):
    def _setUp(self):
        super(JournalWorkerPidFileFixture, self)._setUp()
        # Every pidfile that is created for the JournalPeriodicProcessor
        # worker registers an operation to clean it when the interpreter
        # is about to exit. Tests each have a temporary directory where
        # they work, this directory is deleted after each test. That means
        # that by the time atexit is called the pidfile does not exist anymore
        # and therefore fails with an error. This avoids this problem.
        self.atexit_mock = mock.patch(
            'networking_odl.journal.worker.atexit.register'
        )
        self.atexit_mock.start()


class OpenDaylightPeriodicTaskFixture(fixtures.Fixture):
    def _setUp(self):
        super(OpenDaylightPeriodicTaskFixture, self)._setUp()
        self.task_start_mock = mock.patch.object(
            periodic_task.PeriodicTask, 'start')
        self.task_start_mock.start()


class OpenDaylightPseudoAgentPrePopulateFixture(
        nl_fixture.CallbackRegistryFixture):
    def _setUp(self):
        super(OpenDaylightPseudoAgentPrePopulateFixture, self)._setUp()
        mock.patch.object(
            pseudo_agentdb_binding.PseudoAgentDBBindingPrePopulate,
            'before_port_binding').start()

    # NOTE(yamahata): work around
    # CallbackRegistryFixture._restore causes stopping unstarted patcher
    # bacause some of base classes neutron test cases issue stop_all()
    # with tearDown method
    def _restore(self):
        registry._CALLBACK_MANAGER = self._orig_manager
        # this may cause RuntimeError('stop called on unstarted patcher')
        # due to stop_all called by base test cases
        try:
            self.patcher.stop()
        except RuntimeError:
            pass


class OpenDaylightFullSyncFixture(fixtures.Fixture):
    def _setUp(self):
        super(OpenDaylightFullSyncFixture, self)._setUp()
        self.addCleanup(full_sync.FULL_SYNC_RESOURCES.clear)



# Copyright (c) 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import stevedore

from oslo_config import cfg
from oslo_log import log
from oslo_utils import excutils


LOG = log.getLogger(__name__)


class PortBindingController(object, metaclass=abc.ABCMeta):

    def get_workers(self):
        return []

    @abc.abstractmethod
    def bind_port(self, port_context):
        """Attempt to bind a port.

        :param context: PortContext instance describing the port

        This method is called outside any transaction to attempt to
        establish a port binding using calling mechanism driver. Bindings
        may be created at each of multiple levels of a hierarchical
        network, and are established from the top level downward. At
        each level, the mechanism driver determines whether it can
        bind to any of the network segments in the
        context.segments_to_bind property, based on the value of the
        context.host property, any relevant port or network
        attributes, and its own knowledge of the network topology. At
        the top level, context.segments_to_bind contains the static
        segments of the port's network. At each lower level of
        binding, it contains static or dynamic segments supplied by
        the driver that bound at the level above. If the driver is
        able to complete the binding of the port to any segment in
        context.segments_to_bind, it must call context.set_binding
        with the binding details. If it can partially bind the port,
        it must call context.continue_binding with the network
        segments to be used to bind at the next lower level.
        If the binding results are committed after bind_port returns,
        they will be seen by all mechanism drivers as
        update_port_precommit and update_port_postcommit calls. But if
        some other thread or process concurrently binds or updates the
        port, these binding results will not be committed, and
        update_port_precommit and update_port_postcommit will not be
        called on the mechanism drivers with these results. Because
        binding results can be discarded rather than committed,
        drivers should avoid making persistent state changes in
        bind_port, or else must ensure that such state changes are
        eventually cleaned up.
        Implementing this method explicitly declares the mechanism
        driver as having the intention to bind ports. This is inspected
        by the QoS service to identify the available QoS rules you
        can use with ports.
        """


class PortBindingManager(PortBindingController):
    # At this point, there is no requirement to have multiple
    # port binding controllers at the same time.
    # Stay with single controller until there is a real requirement

    def __init__(self, name, controller):
        self.name = name
        self.controller = controller

    @classmethod
    def create(
            cls, namespace='networking_odl.ml2.port_binding_controllers',
            name=None):

        name = name or cfg.CONF.ml2_odl.port_binding_controller
        ext_mgr = stevedore.named.NamedExtensionManager(
            namespace, [name], invoke_on_load=True)

        assert len(ext_mgr.extensions) == 1, (
            "Wrong port binding controller is specified")

        extension = ext_mgr.extensions[0]
        if isinstance(extension.obj, PortBindingController):
            return cls(extension.name, extension.obj)
        else:
            raise ValueError(_(
                "Port binding controller '%(name)s (%(controller)r)' "
                "doesn't implement PortBindingController interface."),
                {'name': extension.name, 'controller': extension.obj})

    def get_workers(self):
        return self.controller.get_workers()

    def bind_port(self, port_context):
        controller_details = {'name': self.name, 'controller': self.controller}
        try:
            self.controller.bind_port(port_context)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception(
                    "Controller '%(name)s (%(controller)r)' had an error "
                    "when binding port.", controller_details)
        else:
            if port_context._new_bound_segment:
                LOG.info(
                    "Controller '%(name)s (%(controller)r)' has bound port.",
                    controller_details)
            else:
                LOG.debug(
                    "Controller %(name)s (%(controller)r) hasn't bound "
                    "port.", controller_details)



# Copyright 2016 Isaku Yamahata <isaku.yamahata at gmail.com>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add sequence number to journal

Revision ID: 3d560427d776
Revises: 703dbf02afde
Create Date: 2016-08-05 15:50:22.151078

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '3d560427d776'
down_revision = '703dbf02afde'


def upgrade():
    op.create_table(
        'opendaylightjournal_new',
        sa.Column('seqnum', sa.BigInteger(),
                  primary_key=True, autoincrement=True),
        sa.Column('object_type', sa.String(36), nullable=False),
        sa.Column('object_uuid', sa.String(36), nullable=False),
        sa.Column('operation', sa.String(36), nullable=False),
        sa.Column('data', sa.PickleType, nullable=True),
        sa.Column('state',
                  sa.Enum('pending', 'processing', 'failed', 'completed',
                          name='state'),
                  nullable=False, default='pending'),
        sa.Column('retry_count', sa.Integer, default=0),
        sa.Column('created_at', sa.DateTime, default=sa.func.now()),
        sa.Column('last_retried', sa.TIMESTAMP, server_default=sa.func.now(),
                  onupdate=sa.func.now()),
    )



# Copyright (c) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants as n_const
from oslo_log import log
from oslo_serialization import jsonutils

from networking_odl.common import constants as odl_const
from networking_odl.common import utils as odl_utils


LOG = log.getLogger(__name__)


# NOTE(yamahata): As neutron keystone v3 support, tenant_id would be renamed to
# project_id. In order to keep compatibility, populate both
# 'project_id' and 'tenant_id'
# for details refer to
# https://specs.openstack.org/openstack/neutron-specs/specs/newton/moving-to-keystone-v3.html
def _populate_project_id_and_tenant_id(resource_dict):
    # NOTE(yamahata): l3 plugin passes data as dependency_list as python list
    #                 delete_router, delete_floatingip
    if not isinstance(resource_dict, dict):
        return

    project_id = resource_dict.get('project_id',
                                   resource_dict.get('tenant_id'))
    if project_id is not None:
        # NOTE(yamahata): project_id can be ""(empty string)
        resource_dict.setdefault('project_id', project_id)
        resource_dict.setdefault('tenant_id', project_id)


def _filter_unmapped_null(resource_dict, unmapped_keys):
    # NOTE(yamahata): bug work around
    # https://bugs.eclipse.org/bugs/show_bug.cgi?id=475475
    #   Null-value for an unmapped element causes next mapped
    #   collection to contain a null value
    #   JSON: { "unmappedField": null, "mappedCollection": [ "a" ] }
    #
    #   Java Object:
    #   class Root {
    #     Collection<String> mappedCollection = new ArrayList<String>;
    #   }
    #
    #   Result:
    #   Field B contains one element; null
    #
    # TODO(yamahata): update along side with neutron and ODL
    #   add when neutron adds more extensions
    #   delete when ODL neutron northbound supports it
    # TODO(yamahata): do same thing for other resources
    keys_to_del = [key for key in unmapped_keys
                   if resource_dict.get(key) is None]
    if keys_to_del:
        odl_utils.try_del(resource_dict, keys_to_del)


_NETWORK_UNMAPPED_KEYS = ['qos_policy_id']
_SUBNET_UNMAPPED_KEYS = ['segment_id', 'subnetpool_id']
_PORT_UNMAPPED_KEYS = ['binding:profile', 'dns_name',
                       'port_security_enabled', 'qos_policy_id']
_FIP_UNMAPPED_KEYS = ['port_id', 'fixed_ip_address', 'router_id']


def _filter_network_create(network):
    odl_utils.try_del(network, ['status', 'subnets'])
    _filter_unmapped_null(network, _NETWORK_UNMAPPED_KEYS)


def _filter_network_update(network):
    odl_utils.try_del(network, ['id', 'status', 'subnets',
                                'tenant_id', 'project_id'])
    _filter_unmapped_null(network, _NETWORK_UNMAPPED_KEYS)


def _filter_floatingip(fip):
    _filter_unmapped_null(fip, _FIP_UNMAPPED_KEYS)


def _filter_subnet_create(subnet):
    _filter_unmapped_null(subnet, _SUBNET_UNMAPPED_KEYS)


def _filter_subnet_update(subnet):
    odl_utils.try_del(subnet, ['id', 'network_id', 'ip_version', 'cidr',
                               'tenant_id', 'project_id'])
    _filter_unmapped_null(subnet, _SUBNET_UNMAPPED_KEYS)


def _convert_value_to_str(dictionary, key):
    try:
        # use jsonutils to convert unicode & ascii
        dictionary[key] = jsonutils.dumps(dictionary[key])
    except KeyError:
        LOG.warning("key %s is not present in dict %s", key, dictionary)


def _filter_port(port, attributes):
    odl_utils.try_del(port, attributes)
    _filter_unmapped_null(port, _PORT_UNMAPPED_KEYS)
    # ODL excpects binding:profile to be a string, not a dict
    _convert_value_to_str(port, key='binding:profile')


def _filter_port_create(port):
    """Filter out port attributes not required for a create."""
    _filter_port(port, ['status'])


def _filter_port_update(port):
    """Filter out port attributes for an update operation."""
    _filter_port(port, ['network_id', 'id', 'status', 'tenant_id',
                        'project_id'])


def _filter_router_update(router):
    """Filter out attributes for an update operation."""
    odl_utils.try_del(router, ['id', 'tenant_id', 'project_id', 'status'])


# neutron has multiple ICMPv6 names
# https://bugs.launchpad.net/tempest/+bug/1671366
# REVISIT(yamahata): once neutron upstream is fixed to store unified form,
#                    this can be removed.
_ICMPv6_NAMES = (
    n_const.PROTO_NAME_ICMP,
    n_const.PROTO_NAME_IPV6_ICMP,
    n_const.PROTO_NAME_IPV6_ICMP_LEGACY,
)


def _sgrule_scrub_icmpv6_name(sgrule):
    if (sgrule.get('ethertype') == n_const.IPv6 and
            sgrule.get('protocol') in _ICMPv6_NAMES):
        sgrule['protocol'] = n_const.PROTO_NAME_IPV6_ICMP_LEGACY


# ODL neturon northbound knows the following protocol names.
# It's safe to pass those names
_ODL_KNOWN_PROTOCOL_NAMES = (
    n_const.PROTO_NAME_TCP,
    n_const.PROTO_NAME_UDP,
    n_const.PROTO_NAME_ICMP,
    n_const.PROTO_NAME_IPV6_ICMP_LEGACY,
)


def _sgrule_scrub_unknown_protocol_name(protocol):
    """Convert unknown protocol name to actual interger.

    OpenDaylight does't want to keep catching up list of protocol names.
    So networking-odl converts unknown protcol name into integer
    """
    if protocol in _ODL_KNOWN_PROTOCOL_NAMES:
        return protocol
    if protocol in n_const.IP_PROTOCOL_MAP:
        return n_const.IP_PROTOCOL_MAP[protocol]
    return protocol


def _filter_security_group_rule(sg_rule):
    _sgrule_scrub_icmpv6_name(sg_rule)
    if sg_rule.get('protocol'):
        sg_rule['protocol'] = _sgrule_scrub_unknown_protocol_name(
            sg_rule['protocol'])


_FILTER_MAP = {
    (odl_const.ODL_NETWORK, odl_const.ODL_CREATE): _filter_network_create,
    (odl_const.ODL_NETWORK, odl_const.ODL_UPDATE): _filter_network_update,
    (odl_const.ODL_SUBNET, odl_const.ODL_CREATE): _filter_subnet_create,
    (odl_const.ODL_SUBNET, odl_const.ODL_UPDATE): _filter_subnet_update,
    (odl_const.ODL_PORT, odl_const.ODL_CREATE): _filter_port_create,
    (odl_const.ODL_PORT, odl_const.ODL_UPDATE): _filter_port_update,
    (odl_const.ODL_ROUTER, odl_const.ODL_UPDATE): _filter_router_update,
    (odl_const.ODL_SG_RULE, odl_const.ODL_CREATE): _filter_security_group_rule,
    (odl_const.ODL_SG_RULE, odl_const.ODL_UPDATE): _filter_security_group_rule,
    (odl_const.ODL_FLOATINGIP, odl_const.ODL_UPDATE): _filter_floatingip,
}


def filter_for_odl(object_type, operation, data):
    """Filter out the attributed before sending the data to ODL"""
    filter_key = (object_type, operation)
    if filter_key in _FILTER_MAP:
        _FILTER_MAP[filter_key](data)
    _populate_project_id_and_tenant_id(data)



# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import fixture as config_fixture
from oslo_serialization import jsonutils
import requests
import websocket

from networking_odl.common.client import OpenDaylightRestClient as odl_client
from networking_odl.common import websocket_client as wsc
from networking_odl.tests import base


class TestWebsocketClient(base.DietTestCase):
    """Test class for Websocket Client."""

    FAKE_WEBSOCKET_STREAM_NAME_DATA = {
        'output': {
            'stream-name': 'data-change-event-subscription/neutron:neutron/'
            'neutron:hostconfigs/datastore=OPERATIONAL/scope=SUBTREE'
        }}

    INVALID_WEBSOCKET_STREAM_NAME_DATA = {
        'outputs': {
            'stream-name': 'data-change-event-subscription/neutron:neutron/'
            'neutron:hostconfigs/datastore=OPERATIONAL/scope=SUBTREE'
        }}

    FAKE_WEBSOCKET_SUBS_DATA = {
        'location': 'ws://localhost:8185/data-change-event-subscription/'
        'neutron:neutron/neutron:hostconfigs/datastore=OPERATIONAL'
        '/scope=SUBTREE'}

    ODL_URI = "http://localhost:8080/"

    WEBSOCKET_URI = (
        "ws://localhost:8185/" +
        "data-change-event-subscription/neutron:neutron/" +
        "neutron:hostconfigs/datastore=OPERATIONAL/scope=SUBTREE")

    WEBSOCKET_SSL_URI = (
        "wss://localhost:8185/" +
        "data-change-event-subscription/neutron:neutron/" +
        "neutron:hostconfigs/datastore=OPERATIONAL/scope=SUBTREE")

    mock_callback_handler = mock.MagicMock()

    def setUp(self):
        """Setup test."""
        self.useFixture(base.OpenDaylightRestClientFixture())
        mock.patch.object(wsc.OpenDaylightWebsocketClient,
                          'start_odl_websocket_thread').start()
        self.cfg = self.useFixture(config_fixture.Config())
        super(TestWebsocketClient, self).setUp()

        self.mgr = wsc.OpenDaylightWebsocketClient.odl_create_websocket(
            TestWebsocketClient.ODL_URI,
            "restconf/operational/neutron:neutron/hostconfigs",
            wsc.ODL_OPERATIONAL_DATASTORE, wsc.ODL_NOTIFICATION_SCOPE_SUBTREE,
            TestWebsocketClient.mock_callback_handler
        )

    def _get_raised_response(self, status_code):
        response = requests.Response()
        response.status_code = status_code
        return response

    @classmethod
    def _get_mock_request_response(cls, status_code):
        response = mock.Mock(status_code=status_code)
        side_effect = None
        # NOTE(rajivk): requests.codes.bad_request constant value is 400,
        # so it filters requests where client(4XX) or server(5XX) has erred.
        if status_code >= requests.codes.bad_request:
            side_effect = requests.exceptions.HTTPError()

        response.raise_for_status = mock.Mock(side_effect=side_effect)
        return response

    @mock.patch.object(odl_client, 'sendjson')
    def test_subscribe_websocket_sendjson(self, mocked_sendjson):
        request_response = self._get_raised_response(
            requests.codes.unauthorized)

        mocked_sendjson.return_value = request_response
        stream_url = self.mgr._subscribe_websocket()
        self.assertIsNone(stream_url)

        request_response = self._get_raised_response(
            requests.codes.bad_request)

        mocked_sendjson.return_value = request_response
        self.assertRaises(ValueError, self.mgr._subscribe_websocket)

        request_response = self._get_mock_request_response(requests.codes.ok)
        request_response.json = mock.Mock(
            return_value=(TestWebsocketClient.
                          INVALID_WEBSOCKET_STREAM_NAME_DATA))
        mocked_sendjson.return_value = request_response
        self.assertRaises(ValueError, self.mgr._subscribe_websocket)

        request_response = self._get_mock_request_response(requests.codes.ok)
        request_response.json = mock.Mock(return_value={""})
        mocked_sendjson.return_value = request_response
        self.assertIsNone(self.mgr._subscribe_websocket())

    @mock.patch.object(odl_client, 'get')
    def test_subscribe_websocket_get(self, mocked_get):
        request_response = self._get_raised_response(requests.codes.not_found)
        mocked_get.return_value = request_response
        self.assertRaises(ValueError, self.mgr._subscribe_websocket)

        request_response = self._get_raised_response(
            requests.codes.bad_request)

        mocked_get.return_value = request_response
        stream_url = self.mgr._subscribe_websocket()
        self.assertIsNone(stream_url)

        request_response = self._get_raised_response(
            requests.codes.unauthorized)

        mocked_get.return_value = request_response
        stream_url = self.mgr._subscribe_websocket()
        self.assertIsNone(stream_url)

    @mock.patch.object(odl_client, 'sendjson')
    @mock.patch.object(odl_client, 'get')
    def test_subscribe_websocket(self, mocked_get, mocked_sendjson):
        request_response = self._get_mock_request_response(requests.codes.ok)
        request_response.json = mock.Mock(
            return_value=TestWebsocketClient.FAKE_WEBSOCKET_STREAM_NAME_DATA)
        mocked_sendjson.return_value = request_response

        request_response = self._get_mock_request_response(requests.codes.ok)
        request_response.headers = TestWebsocketClient.FAKE_WEBSOCKET_SUBS_DATA
        mocked_get.return_value = request_response
        stream_url = self.mgr._subscribe_websocket()

        self.assertEqual(TestWebsocketClient.WEBSOCKET_URI, stream_url)

    @mock.patch.object(websocket, 'create_connection')
    def test_create_connection(self, mock_create_connection):
        mock_create_connection.return_value = None
        return_value = self.mgr._socket_create_connection("localhost")
        self.assertIsNone(return_value)

    @mock.patch.object(websocket, 'create_connection',
                       side_effect=Exception("something went wrong"))
    def test_create_connection_handles_exception(self, mock_create_connection):
        self.assertIsNone(self.mgr._socket_create_connection("localhost"))

    def test_websocket_connect(self):
        self.mgr._subscribe_websocket = mock.MagicMock(
            return_value=TestWebsocketClient.WEBSOCKET_URI)
        self.mgr._socket_create_connection = mock.MagicMock(return_value=True)
        self.mgr._connect_ws()
        self.mgr._socket_create_connection.assert_called_with(
            TestWebsocketClient.WEBSOCKET_URI)

    def test_websocket_connect_ssl(self):
        self.mgr._subscribe_websocket = mock.MagicMock(
            return_value=TestWebsocketClient.WEBSOCKET_SSL_URI)
        self.mgr._socket_create_connection = mock.MagicMock(return_value=True)
        self.mgr._connect_ws()
        self.mgr._socket_create_connection.assert_called_with(
            TestWebsocketClient.WEBSOCKET_SSL_URI)

    def test_websocket_connect_ssl_negative_uri(self):
        self.mgr._subscribe_websocket = mock.MagicMock(
            return_value=TestWebsocketClient.WEBSOCKET_URI)
        self.mgr._socket_create_connection = mock.MagicMock(return_value=True)
        self.mgr.odl_rest_client.url = self.mgr.odl_rest_client.url.replace(
            'http:', 'https:')
        self.mgr._connect_ws()
        self.mgr._socket_create_connection.assert_called_with(
            TestWebsocketClient.WEBSOCKET_SSL_URI)

    def test_run_websocket_thread(self):
        self.mgr._connect_ws = mock.MagicMock(return_value=None)
        self.cfg.config(restconf_poll_interval=0, group='ml2_odl')
        self.mgr.run_websocket_thread(True)
        assert self.mgr._connect_ws.call_count == 1

        self.mgr.set_exit_flag(False)
        self.mgr._connect_ws = mock.MagicMock(return_value=1)
        with mock.patch.object(wsc, 'LOG') as mock_log:
            self.mgr.run_websocket_thread(True)
            self.assertTrue(mock_log.error.called)

        self.mgr.set_exit_flag(False)
        ws = mock.MagicMock()
        ws.recv.return_value = None
        self.mgr._connect_ws = mock.MagicMock(return_value=ws)
        self.mgr._close_ws = mock.MagicMock(return_value=None)
        with mock.patch.object(wsc, 'LOG') as mock_log:
            self.mgr.run_websocket_thread(True)
            self.assertTrue(mock_log.warning.called)

        self.mgr.set_exit_flag(False)
        ws = mock.MagicMock()
        ws.recv.return_value = "Test Data"
        self.mgr._connect_ws = mock.MagicMock(return_value=ws)
        self.mgr._close_ws = mock.MagicMock(return_value=None)
        self.mgr.run_websocket_thread(True)
        TestWebsocketClient.mock_callback_handler.assert_called_once()


class TestEventDataParser(base.DietTestCase):
    """Test class for Websocket Client."""

    # test data port status payload
    sample_port_status_payload = """{"notification":
        {"xmlns":"urn:ietf:params:xml:ns:netconf:notification:1.0",
         "data-changed-notification": { "xmlns":
             "urn:opendaylight:params:xml:ns:yang:controller:md:sal:remote",
             "data-change-event":
                 [{"path":
                "/neutron:neutron/neutron:ports/neutron:port\
                [neutron:uuid='a51e439f-4d02-4e76-9b0d-08f6c08855dd']\
                /neutron:uuid",
                "data":{"uuid":{"xmlns":"urn:opendaylight:neutron",
                        "content":"a51e439f-4d02-4e76-9b0d-08f6c08855dd"}},
                "operation":"created"},
                {"path":
                "/neutron:neutron/neutron:ports/neutron:port\
                [neutron:uuid='a51e439f-4d02-4e76-9b0d-08f6c08855dd']\
                /neutron:status",
                "data":{"status":{"xmlns":"urn:opendaylight:neutron",
                        "content":"ACTIVE"}},
                 "operation":"created"}
                  ]},
            "eventTime":"2017-03-23T09:28:55.379-07:00"}}"""

    sample_port_status_payload_one_item = """{"notification":
        {"xmlns": "urn:ietf:params:xml:ns:netconf:notification:1.0",
        "data-changed-notification": {
         "data-change-event": {
            "data": { "status": {
                        "content": "ACTIVE",
                        "xmlns": "urn:opendaylight:neutron"
            }},
            "operation": "updated",
            "path": "/neutron:neutron/neutron:ports/neutron:port\
            [neutron:uuid='d6e6335d-9568-4949-aef1-4107e34c5f28']\
            /neutron:status"
            },
            "xmlns": "urn:opendaylight:params:xml:ns:yang:controller:md:\
            sal:remote"
        },
        "eventTime": "2017-02-22T02:27:32+02:00" }}"""

    def setUp(self):
        """Setup test."""
        super(TestEventDataParser, self).setUp()

    def test_get_item_port_status_payload(self):
        sample = jsonutils.loads(self.sample_port_status_payload)
        expected_events = (sample
                           [wsc.EventDataParser.NOTIFICATION_TAG]
                           [wsc.EventDataParser.DC_NOTIFICATION_TAG]
                           [wsc.EventDataParser.DC_EVENT_TAG])
        event_0 = expected_events[0]
        event = wsc.EventDataParser.get_item(self.sample_port_status_payload)
        operation, path, data = next(event).get_fields()

        self.assertEqual(event_0.get('operation'), operation)
        self.assertEqual(event_0.get('path'), path)
        self.assertEqual(event_0.get('data'), data)

        uuid = wsc.EventDataParser.extract_field(path, "neutron:uuid")
        self.assertEqual("'a51e439f-4d02-4e76-9b0d-08f6c08855dd'", uuid)

        uuid = wsc.EventDataParser.extract_field(path, "invalidkey")
        self.assertIsNone(uuid)

    def test_get_item_port_status_payload_one_item(self):
        sample = jsonutils.loads(self.sample_port_status_payload_one_item)
        expected_events = (sample
                           [wsc.EventDataParser.NOTIFICATION_TAG]
                           [wsc.EventDataParser.DC_NOTIFICATION_TAG]
                           [wsc.EventDataParser.DC_EVENT_TAG])
        event = (wsc.EventDataParser.
                 get_item(self.sample_port_status_payload_one_item))
        operation, path, data = next(event).get_fields()

        self.assertEqual(expected_events.get('operation'), operation)
        self.assertEqual(expected_events.get('path'), path)
        self.assertEqual(expected_events.get('data'), data)

        uuid = wsc.EventDataParser.extract_field(path, "neutron:uuid")
        self.assertEqual("'d6e6335d-9568-4949-aef1-4107e34c5f28'", uuid)



#
# Copyright (C) 2016 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

from unittest import mock

from neutron_lib.db import api as db_api
from neutron_lib import exceptions as nexc
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory

from networking_odl.common import constants as odl_const
from networking_odl.common import exceptions
from networking_odl.db import db
from networking_odl.journal import full_sync
from networking_odl.journal import recovery
from networking_odl.l3 import l3_odl_v2
from networking_odl.ml2 import mech_driver_v2
from networking_odl.tests import base
from networking_odl.tests.unit.db import test_db
from networking_odl.tests.unit.journal import helper
from networking_odl.tests.unit import test_base_db


class RecoveryTestCase(test_base_db.ODLBaseDbTestCase):
    def setUp(self):
        self.useFixture(
            base.OpenDaylightRestClientGlobalFixture(recovery._CLIENT))
        super(RecoveryTestCase, self).setUp()
        self._CLIENT = recovery._CLIENT.get_client()
        self.addCleanup(self.clean_registered_resources)

    @staticmethod
    def clean_registered_resources():
        full_sync.ALL_RESOURCES = {}

    def _mock_resource(self, plugin, resource_type):
        mock_resource = mock.MagicMock()
        get_func = getattr(plugin, 'get_{}'.format(resource_type))
        get_func.return_value = mock_resource
        return mock_resource

    def _mock_row(self, resource_type):
        return mock.MagicMock(object_type=resource_type)

    def _test__get_latest_resource(self, plugin, resource_type):
        l2 = mech_driver_v2.OpenDaylightMechanismDriver.RESOURCES
        full_sync.ALL_RESOURCES[plugin_constants.CORE] = l2
        mock_resource = self._mock_resource(plugin, resource_type)
        mock_row = self._mock_row(resource_type)

        resource = recovery._get_latest_resource(self.db_context.session,
                                                 mock_row)
        self.assertEqual(mock_resource, resource)

    @mock.patch.object(directory, 'get_plugin')
    def test__get_latest_resource_l2(self, plugin_mock):
        for resource_type in (
                mech_driver_v2.OpenDaylightMechanismDriver.RESOURCES):

            plugin = plugin_mock.return_value
            self._test__get_latest_resource(plugin, resource_type)

    @mock.patch.object(directory, 'get_plugin')
    def test__get_latest_resource_l3(self, plugin_mock):
        full_sync.ALL_RESOURCES[plugin_constants.L3] = l3_odl_v2.L3_RESOURCES
        for resource_type in l3_odl_v2.L3_RESOURCES:
            plugin = plugin_mock.return_value
            self._test__get_latest_resource(plugin, resource_type)

    def test__get_latest_resource_unsupported(self):
        mock_row = self._mock_row('aaa')
        self.assertRaises(
            exceptions.UnsupportedResourceType, recovery._get_latest_resource,
            self.db_context.session, mock_row)

    @mock.patch.object(directory, 'get_plugin')
    def test__get_latest_resource_none(self, plugin_mock):
        plugin_mock.return_value.get_network.side_effect = nexc.NotFound()
        l2 = mech_driver_v2.OpenDaylightMechanismDriver.RESOURCES
        full_sync.ALL_RESOURCES[plugin_constants.CORE] = l2

        mock_row = self._mock_row(odl_const.ODL_NETWORK)
        self.assertRaises(
            nexc.NotFound, recovery._get_latest_resource,
            self.db_context.session, mock_row)

    def test_journal_recovery_retries_exceptions(self):
        with db_api.CONTEXT_WRITER.using(self.db_context):
            db.create_pending_row(self.db_context, odl_const.ODL_NETWORK,
                                  'id', odl_const.ODL_DELETE, {})
            created_row = db.get_all_db_rows(self.db_context)[0]
            db.update_db_row_state(self.db_context, created_row,
                                   odl_const.FAILED)
        with mock.patch.object(db, 'update_db_row_state') as m:
            self._test_retry_exceptions(recovery.journal_recovery, m, False)

    def test_journal_recovery_no_rows(self):
        recovery.journal_recovery(self.db_context)
        self.assertFalse(self._CLIENT.get_resource.called)

    @test_db.in_session
    def _test_recovery(self, operation, odl_resource, expected_state):
        db.create_pending_row(self.db_context, odl_const.ODL_NETWORK,
                              'id', operation, {})
        created_row = db.get_all_db_rows(self.db_context)[0]
        db.update_db_row_state(self.db_context, created_row, odl_const.FAILED)

        self._CLIENT.get_resource.return_value = odl_resource

        recovery.journal_recovery(self.db_context)

        if expected_state is None:
            completed_rows = db.get_all_db_rows_by_state(
                self.db_context, odl_const.COMPLETED)
            self.assertEqual([], completed_rows)
        else:
            row = db.get_all_db_rows_by_state(self.db_context,
                                              expected_state)[0]
            self.assertEqual(created_row['seqnum'], row['seqnum'])

        return created_row

    def _disable_retention(self):
        self.cfg.config(completed_rows_retention=0, group='ml2_odl')

    def test_journal_recovery_handles_failure_quietly(self):
        class TestException(Exception):
            pass

        self._CLIENT.get_resource.side_effect = TestException('')
        self._test_recovery(
            odl_const.ODL_DELETE, None, odl_const.FAILED)

    def test_journal_recovery_deleted_row_not_in_odl(self):
        self._test_recovery(odl_const.ODL_DELETE, None, odl_const.COMPLETED)

    def test_journal_recovery_deleted_row_not_in_odl_purged(self):
        self._disable_retention()
        self._test_recovery(odl_const.ODL_DELETE, None, None)

    def test_journal_recovery_created_row_exists_in_odl(self):
        self._test_recovery(odl_const.ODL_CREATE, {}, odl_const.COMPLETED)

    def test_journal_recovery_created_row_exists_in_odl_purged(self):
        self._disable_retention()
        self._test_recovery(odl_const.ODL_CREATE, {}, None)

    def test_journal_recovery_deleted_row_exists_in_odl(self):
        self._test_recovery(odl_const.ODL_DELETE, {}, odl_const.PENDING)

    @mock.patch.object(recovery, '_get_latest_resource')
    def _test_recovery_creates_operation(
            self, operation, resource, odl_resource, expected_operation,
            recovery_mock):
        if resource is not None:
            recovery_mock.return_value = resource
        else:
            recovery_mock.side_effect = nexc.NotFound
        original_row = self._test_recovery(
            operation, odl_resource, odl_const.COMPLETED)

        pending_row = db.get_all_db_rows_by_state(
            self.db_context, odl_const.PENDING)[0]
        self.assertEqual(expected_operation, pending_row['operation'])
        self.assertEqual(original_row['object_type'],
                         pending_row['object_type'])
        self.assertEqual(original_row['object_uuid'],
                         pending_row['object_uuid'])

    def test_recovery_created_row_not_in_odl(self):
        self._test_recovery_creates_operation(
            odl_const.ODL_CREATE, {}, None, odl_const.ODL_CREATE)

    def test_recovery_updated_row_not_in_odl(self):
        self._test_recovery_creates_operation(
            odl_const.ODL_UPDATE, {}, None, odl_const.ODL_CREATE)

    def test_recovery_updated_resource_missing_but_exists_in_odl(self):
        self._test_recovery_creates_operation(
            odl_const.ODL_UPDATE, None, {}, odl_const.ODL_DELETE)

    @mock.patch.object(recovery, '_get_latest_resource')
    def test_recovery_created_resource_missing_and_not_in_odl(self, rmock):
        rmock.side_effect = nexc.NotFound
        self._test_recovery(odl_const.ODL_CREATE, None, odl_const.COMPLETED)

    @mock.patch.object(recovery, '_get_latest_resource')
    def test_recovery_created_resource_missing_and_not_in_odl_purged(
            self, rmock):
        rmock.side_effect = nexc.NotFound
        self._disable_retention()
        self._test_recovery(odl_const.ODL_CREATE, None, None)

    @mock.patch.object(recovery, '_get_latest_resource')
    def test_recovery_updated_resource_missing_and_not_in_odl(self, rmock):
        rmock.side_effect = nexc.NotFound
        self._test_recovery(odl_const.ODL_UPDATE, None, odl_const.COMPLETED)

    @mock.patch.object(recovery, '_get_latest_resource')
    def test_recovery_updated_resource_missing_and_not_in_odl_purged(
            self, rmock):
        rmock.side_effect = nexc.NotFound
        self._disable_retention()
        self._test_recovery(odl_const.ODL_UPDATE, None, None)

    def _test_get_latest_resource(self, resource_type):
        # Drivers needs to be initialized to register resources for recovery
        # and full sync mechasnim.
        helper.TestDriver()
        directory.add_plugin(helper.TEST_PLUGIN, helper.TestPlugin())
        self.addCleanup(directory.add_plugin, helper.TEST_PLUGIN, None)
        return db.create_pending_row(self.db_context, resource_type,
                                     'id', odl_const.ODL_DELETE, {})

    def test_get_latest_resource(self):
        row = self._test_get_latest_resource(helper.TEST_RESOURCE1)
        plugin = directory.get_plugin(helper.TEST_PLUGIN)
        resource = recovery.get_latest_resource(self.db_context, row)
        self.assertDictEqual(resource,
                             plugin.get_test_resource1(self.db_context, 'id'))

    def test_get_unsupported_latest_resource(self):
        row = self._test_get_latest_resource(helper.TEST_RESOURCE1)
        row.object_type = helper.INVALID_RESOURCE
        self.assertRaises(exceptions.UnsupportedResourceType,
                          recovery.get_latest_resource,
                          self.db_context, row)



# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import testscenarios

from neutron_lib import constants as n_const
from neutron_lib.plugins import directory

from networking_odl.common import constants as odl_const
from networking_odl.dhcp import odl_dhcp_driver_base as driver_base
from networking_odl.ml2 import mech_driver_v2
from networking_odl.tests import base as odl_base
from networking_odl.tests.unit import base_v2

from oslo_config import cfg
from oslo_utils import uuidutils

# Required to generate tests from scenarios. Not compatible with nose.
load_tests = testscenarios.load_tests_apply_scenarios
ODL_TENANT_ID = uuidutils.generate_uuid()
cfg.CONF.import_group('ml2_odl', 'networking_odl.common.config')


class OdlDhcpDriverTestBase(base_v2.OpenDaylightConfigBase):

    def setUp(self):
        self.useFixture(odl_base.OpenDaylightFeaturesFixture())
        self.useFixture(odl_base.OpenDaylightJournalThreadFixture())
        self.useFixture(odl_base.OpenDaylightPseudoAgentPrePopulateFixture())
        super(OdlDhcpDriverTestBase, self).setUp()

    def get_network_and_subnet_context(self, cidr, dhcp_flag, create_subnet,
                                       create_network, ipv4=True):
        data = {}
        network_id = uuidutils.generate_uuid()
        subnet_id = uuidutils.generate_uuid()
        plugin = directory.get_plugin()
        data['network_id'] = network_id
        data['subnet_id'] = subnet_id
        data['context'] = self.context
        data['plugin'] = plugin
        network, network_context = \
            self.get_network_context(network_id, create_network, ipv4)
        if create_network:
            data['network_context'] = network_context
        data['network'] = network
        subnet, subnet_context = \
            self.get_subnet_context(network_id, subnet_id, cidr,
                                    dhcp_flag, create_subnet, ipv4)
        if create_subnet:
            data['subnet_context'] = subnet_context
        data['subnet'] = subnet
        return data

    def get_subnet_context(self, network_id, subnet_id, cidr,
                           dhcp_flag, create_subnet, ipv4=True):
        if ipv4:
            index = cidr.rfind('.') + 1
            ip_range = cidr[:index]
            cidr_end = ip_range + str(254)
            ipv6_ramode = None
            ipv6_addmode = None
            ipversion = 4
        else:
            index = cidr.rfind(':') + 1
            ip_range = cidr[:index]
            cidr_end = cidr[:index - 1] + 'ffff:ffff:ffff:fffe'
            ipv6_ramode = 'slaac'
            ipv6_addmode = 'slaac'
            ipversion = 6
        current = {'ipv6_ra_mode': ipv6_ramode,
                   'allocation_pools': [{'start': ip_range + str(2),
                                         'end': cidr_end}],
                   'host_routes': [],
                   'ipv6_address_mode': ipv6_addmode,
                   'cidr': cidr,
                   'id': subnet_id,
                   'name': '',
                   'enable_dhcp': dhcp_flag,
                   'network_id': network_id,
                   'tenant_id': ODL_TENANT_ID,
                   'project_id': ODL_TENANT_ID,
                   'dns_nameservers': [],
                   'gateway_ip': ip_range + str(1),
                   'ip_version': ipversion,
                   'shared': False}
        subnet = {'subnet': AttributeDict(current)}
        if create_subnet:
            plugin = directory.get_plugin()
            result, subnet_context = plugin._create_subnet_db(self.context,
                                                              subnet)
            return subnet, subnet_context
        else:
            return subnet

    def get_network_context(self, network_id, create_network, ipv4=True):
        netwrk = 'netv4'
        if not ipv4:
            netwrk = 'netv6'
        current = {'status': 'ACTIVE',
                   'subnets': [],
                   'name': netwrk,
                   'provider:physical_network': None,
                   'admin_state_up': True,
                   'tenant_id': ODL_TENANT_ID,
                   'project_id': ODL_TENANT_ID,
                   'provider:network_type': 'local',
                   'router:external': False,
                   'shared': False,
                   'id': network_id,
                   'provider:segmentation_id': None}
        network = {'network': AttributeDict(current)}
        if create_network:
            plugin = directory.get_plugin()
            result, network_context = plugin._create_network_db(
                self.context, network)
            return [network, network_context]
        return network

    def get_port_id(self, plugin, plugin_context, network_id, subnet_id):

        device_id = odl_const.ODL_DEVICE_ID_START + '-' + subnet_id
        filters = {
            'network_id': [network_id],
            'device_id': [device_id],
            'device_owner': [n_const.DEVICE_OWNER_DHCP]
        }
        ports = plugin.get_ports(plugin_context, filters=filters)
        if ports:
            port = ports[0]
            return port['id']


class OdlDhcpDriverBaseTestCase(OdlDhcpDriverTestBase):

    def setUp(self):
        super(OdlDhcpDriverBaseTestCase, self).setUp()

    def test_dhcp_driver_not_loaded_without_flag(self):
        mech = mech_driver_v2.OpenDaylightMechanismDriver()
        mech.initialize()
        args = [mech, 'dhcp_driver']
        self.assertRaises(AttributeError, getattr, *args)

    def test_dhcp_port_create(self):
        dhcp_driver = driver_base.OdlDhcpDriverBase()

        data = self.get_network_and_subnet_context('10.0.10.0/24', True, True,
                                                   True)

        dhcp_driver.create_or_delete_dhcp_port(data['subnet_context'])

        port = self.get_port_id(data['plugin'],
                                data['context'],
                                data['network_id'],
                                data['subnet_id'])
        self.assertIsNotNone(port)

    def test_dhcp_port_create_v6network(self):
        dhcp_driver = driver_base.OdlDhcpDriverBase()

        data = self.get_network_and_subnet_context('2001:db8:abcd:0012::0/64',
                                                   True, True, True, False)
        dhcp_driver.create_or_delete_dhcp_port(data['subnet_context'])

        port = self.get_port_id(data['plugin'],
                                data['context'],
                                data['network_id'],
                                data['subnet_id'])
        self.assertIsNone(port)

    def test_dhcp_create_without_dhcp_flag(self):
        dhcp_driver = driver_base.OdlDhcpDriverBase()
        data = self.get_network_and_subnet_context('10.0.20.0/24', False, True,
                                                   True)

        dhcp_driver.create_or_delete_dhcp_port(data['subnet_context'])

        port = self.get_port_id(data['plugin'],
                                data['context'],
                                data['network_id'],
                                data['subnet_id'])
        self.assertIsNone(port)

    def test_dhcp_port_create_with_multiple_create_request(self):
        dhcp_driver = driver_base.OdlDhcpDriverBase()
        data = self.get_network_and_subnet_context('10.0.30.0/24', True, True,
                                                   True)

        dhcp_driver.create_or_delete_dhcp_port(data['subnet_context'])
        dhcp_driver.create_or_delete_dhcp_port(data['subnet_context'])
        # If there are multiple ports will one_or_none wiill throw error
        # MultipleResultsFound
        port = self.get_port_id(data['plugin'],
                                data['context'],
                                data['network_id'],
                                data['subnet_id'])
        self.assertIsNotNone(port)

    def test_dhcp_update_from_disable_to_enable(self):
        dhcp_driver = driver_base.OdlDhcpDriverBase()
        data = self.get_network_and_subnet_context('10.0.40.0/24', False, True,
                                                   True)
        subnet_context = data['subnet_context']
        dhcp_driver.create_or_delete_dhcp_port(subnet_context)
        subnet_context.current['enable_dhcp'] = True
        dhcp_driver.create_or_delete_dhcp_port(subnet_context)

        port = self.get_port_id(data['plugin'],
                                data['context'],
                                data['network_id'],
                                data['subnet_id'])
        self.assertIsNotNone(port)

    def test_dhcp_update_from_enable_to_enable(self):
        dhcp_driver = driver_base.OdlDhcpDriverBase()
        data = self.get_network_and_subnet_context('10.0.50.0/24', True, True,
                                                   True)
        subnet_context = data['subnet_context']
        dhcp_driver.create_or_delete_dhcp_port(subnet_context)
        subnet_context.current['enable_dhcp'] = True
        dhcp_driver.create_or_delete_dhcp_port(subnet_context)

        port = self.get_port_id(data['plugin'],
                                data['context'],
                                data['network_id'],
                                data['subnet_id'])
        self.assertIsNotNone(port)

    def test_dhcp_update_from_enable_to_disable(self):
        dhcp_driver = driver_base.OdlDhcpDriverBase()
        data = self.get_network_and_subnet_context('10.0.60.0/24', True, True,
                                                   True)
        subnet_context = data['subnet_context']
        dhcp_driver.create_or_delete_dhcp_port(subnet_context)
        subnet_context.current['enable_dhcp'] = False
        dhcp_driver.create_or_delete_dhcp_port(subnet_context)

        port = self.get_port_id(data['plugin'],
                                data['context'],
                                data['network_id'],
                                data['subnet_id'])
        self.assertIsNone(port)

    def test_dhcp_update_from_disable_to_disable(self):
        dhcp_driver = driver_base.OdlDhcpDriverBase()
        data = self.get_network_and_subnet_context('10.0.70.0/24', False, True,
                                                   True)
        subnet_context = data['subnet_context']
        dhcp_driver.create_or_delete_dhcp_port(subnet_context)
        subnet_context.current['enable_dhcp'] = False
        dhcp_driver.create_or_delete_dhcp_port(subnet_context)

        port = self.get_port_id(data['plugin'],
                                data['context'],
                                data['network_id'],
                                data['subnet_id'])
        self.assertIsNone(port)

    def test_dhcp_delete_when_dhcp_enabled(self):
        dhcp_driver = driver_base.OdlDhcpDriverBase()
        data = self.get_network_and_subnet_context('10.0.80.0/24', True, True,
                                                   True)
        subnet_context = data['subnet_context']
        dhcp_driver.create_or_delete_dhcp_port(subnet_context)
        subnet_context.current['enable_dhcp'] = False
        dhcp_driver.create_or_delete_dhcp_port(subnet_context)
        port = self.get_port_id(data['plugin'],
                                data['context'],
                                data['network_id'],
                                data['subnet_id'])
        self.assertIsNone(port)

    def test_dhcp_delete_when_dhcp_delete(self):
        dhcp_driver = driver_base.OdlDhcpDriverBase()
        data = self.get_network_and_subnet_context('10.0.90.0/24', False, True,
                                                   True)
        subnet_context = data['subnet_context']
        dhcp_driver.create_or_delete_dhcp_port(subnet_context)
        dhcp_driver.create_or_delete_dhcp_port(subnet_context)
        port = self.get_port_id(data['plugin'],
                                data['context'],
                                data['network_id'],
                                data['subnet_id'])
        self.assertIsNone(port)


class AttributeDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttributeDict, self).__init__(*args, **kwargs)
        self.__dict__ = self



#
# Copyright (C) 2016 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

import functools

from neutron.tests.unit.extensions import test_securitygroup
from neutron.tests.unit.plugins.ml2 import test_plugin

from networking_odl.common import constants as odl_const
from networking_odl.tests.functional import base


class _DriverTest(base.OdlTestsBase):

    def test_network_create(self):
        with self.network() as network:
            self.assert_resource_created(odl_const.ODL_NETWORK, network)

    def test_network_update(self):
        with self.network() as network:
            self.resource_update_test(odl_const.ODL_NETWORK, network)

    def test_network_delete(self):
        with self.network() as network:
            self.resource_delete_test(odl_const.ODL_NETWORK, network)

    def test_subnet_create(self):
        with self.network() as network:
            with self.subnet(network=network) as subnet:
                self.assert_resource_created(odl_const.ODL_SUBNET, subnet)

    def test_subnet_update(self):
        with self.network() as network:
            with self.subnet(network=network) as subnet:
                self.resource_update_test(odl_const.ODL_SUBNET, subnet)

    def test_subnet_delete(self):
        with self.network() as network:
            with self.subnet(network=network) as subnet:
                self.resource_delete_test(odl_const.ODL_SUBNET, subnet)

    def test_port_create(self):
        with self.network() as network:
            with self.subnet(network=network) as subnet:
                with self.port(subnet=subnet) as port:
                    self.assert_resource_created(odl_const.ODL_PORT, port)

    def test_port_update(self):
        with self.network() as network:
            with self.subnet(network=network) as subnet:
                with self.port(subnet=subnet) as port:
                    self.resource_update_test(odl_const.ODL_PORT, port)

    def test_port_delete(self):
        with self.network() as network:
            with self.subnet(network=network) as subnet:
                with self.port(subnet=subnet) as port:
                    self.resource_delete_test(odl_const.ODL_PORT, port)


class _DriverSecGroupsTests(base.OdlTestsBase):

    # Override default behavior so that extension manager is used, otherwise
    # we can't test security groups.
    def setup_parent(self):
        """Perform parent setup with the common plugin configuration class."""
        ext_mgr = (
            test_securitygroup.SecurityGroupTestExtensionManager())
        # Ensure that the parent setup can be called without arguments
        # by the common configuration setUp.
        parent_setup = functools.partial(
            super(test_plugin.Ml2PluginV2TestCase, self).setUp,
            plugin=test_plugin.PLUGIN_NAME,
            ext_mgr=ext_mgr,
        )
        self.useFixture(test_plugin.Ml2ConfFixture(parent_setup))

    def test_security_group_create(self):
        with self.security_group() as sg:
            self.assert_resource_created(odl_const.ODL_SG, sg)

    def test_security_group_update(self):
        with self.security_group() as sg:
            self.resource_update_test(odl_const.ODL_SG, sg)

    def test_security_group_delete(self):
        with self.security_group() as sg:
            self.resource_delete_test(odl_const.ODL_SG, sg)

    def test_security_group_rule_create(self):
        with self.security_group() as sg:
            sg_id = sg[odl_const.ODL_SG]['id']
            with self.security_group_rule(security_group_id=sg_id) as sg_rule:
                self.assert_resource_created(odl_const.ODL_SG_RULE, sg_rule)

    def test_security_group_rule_delete(self):
        with self.security_group() as sg:
            sg_id = sg[odl_const.ODL_SG]['id']
            with self.security_group_rule(security_group_id=sg_id) as sg_rule:
                self.resource_delete_test(odl_const.ODL_SG_RULE, sg_rule)


class TestV2Driver(base.V2DriverAdjustment, _DriverTest,
                   test_plugin.Ml2PluginV2TestCase):
    _mechanism_drivers = ['opendaylight_v2']


class TestV2DriverSecGroups(base.V2DriverAdjustment, _DriverSecGroupsTests,
                            test_securitygroup.SecurityGroupsTestCase,
                            test_plugin.Ml2PluginV2TestCase):
    _mechanism_drivers = ['opendaylight_v2']



# Copyright 2017 Red Hat Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Add journal dependencies table

Revision ID: 0472f56ff2fb
Revises: 43af357fd638
Create Date: 2017-04-02 11:02:01.622548

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '0472f56ff2fb'
down_revision = '43af357fd638'


def upgrade():
    op.create_table(
        'opendaylight_journal_deps',
        sa.Column('depends_on', sa.BigInteger(),
                  sa.ForeignKey('opendaylightjournal.seqnum',
                                ondelete='CASCADE'),
                  primary_key=True),
        sa.Column('dependent', sa.BigInteger(),
                  sa.ForeignKey('opendaylightjournal.seqnum',
                                ondelete='CASCADE'),
                  primary_key=True))



# Copyright (c) 2017 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_odl.journal import base_driver


TEST_UUID = 'bd8db3a8-2b30-4083-a8b3-b3fd46401142'
TEST_PLUGIN = 'test_plugin'
TEST_RESOURCE1 = 'test_resource1'
TEST_RESOURCE2 = 'test_resource2'
TEST_RESOURCE1_SUFFIX = 'test_resource1s'
TEST_RESOURCE2_SUFFIX = 'test_resource2s'
INVALID_RESOURCE = 'invalid_resource'
INVALID_PLUGIN = 'invalid_plugin'
INVALID_METHOD = 'invalid_method_name'


class TestPlugin(object):
    def get_test_resource1s(self, context):
        return [{'id': 'test_id1'}, {'id': 'test_id2'}]

    def get_test_resource2s(self, context):
        return [{'id': 'test_id3'}, {'id': 'test_id4'}]

    def get_test_resource1(self, context, id_):
        return {'id': id_}


class TestDriver(base_driver.ResourceBaseDriver):
    RESOURCES = {
        TEST_RESOURCE1: TEST_RESOURCE1_SUFFIX,
        TEST_RESOURCE2: TEST_RESOURCE2_SUFFIX
    }
    plugin_type = TEST_PLUGIN

    def __init__(self):
        super(TestDriver, self).__init__()



# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_odl.common import postcommit
from neutron.tests import base


class BaseTest(object):
    def create_resource1_postcommit(self):
        pass

    update_resource1_postcommit = create_resource1_postcommit
    delete_resource1_postcommit = create_resource1_postcommit
    update_resource2_postcommit = create_resource1_postcommit
    delete_resource2_postcommit = create_resource1_postcommit
    create_resource2_postcommit = create_resource1_postcommit


class TestPostCommit(base.DietTestCase):
    def _get_class(self, *args):
        @postcommit.add_postcommit(*args)
        class TestClass(BaseTest):
            pass

        return TestClass

    def _get_methods_name(self, resources):
        ops = ['create', 'update', 'delete']
        m_names = [op + '_' + resource + '_postcommit' for op in ops
                   for resource in resources]

        return m_names

    def test_with_one_resource(self):
        cls = self._get_class('resource1')
        m_names = self._get_methods_name(['resource1'])
        for m_name in m_names:
            self.assertEqual(m_name, getattr(cls, m_name).__name__)

    def test_with_two_resource(self):
        cls = self._get_class('resource1', 'resource2')
        m_names = self._get_methods_name(['resource1', 'resource2'])
        for m_name in m_names:
            self.assertEqual(m_name, getattr(cls, m_name).__name__)

    def test_with_two_resource_create_defined_for_one(self):
        m_names = self._get_methods_name(['resource1', 'resource2'])

        @postcommit.add_postcommit('resource1', 'resource2')
        class TestClass(BaseTest):
            def create_resource1_postcommit(self):
                pass

            create_resource1_postcommit.__name__ = 'test_method'

        for m_name in m_names[1:]:
            self.assertEqual(m_name, getattr(TestClass, m_name).__name__)

        self.assertEqual('test_method',
                         getattr(TestClass, m_names[0]).__name__)



# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import itertools
import time

from oslo_config import cfg
from oslo_log import log
from oslo_serialization import jsonutils
from requests import exceptions

from networking_odl.common import client as odl_client
from networking_odl.common import utils


cfg.CONF.import_group('ml2_odl', 'networking_odl.common.config')
LOG = log.getLogger(__name__)

OPERATIONAL_PORT_STATUS = 'operational-port-status'

EMPTY_FEATURES = {}

feature_configs = copy.copy(EMPTY_FEATURES)


def init():
    '''initialize odl_features.

    Initialize odl_features. Try first from configuration and then try pulling
    via rest call from ODL.
    '''

    global feature_configs
    feature_configs = None

    if cfg.CONF.ml2_odl.odl_features_json is not None:
        json = jsonutils.loads(cfg.CONF.ml2_odl.odl_features_json)
        feature_configs = _load_features(json)
        return

    if cfg.CONF.ml2_odl.odl_features is not None:
        feature_configs = {feature: '' for feature
                           in cfg.CONF.ml2_odl.odl_features}
        return

    wait_interval = cfg.CONF.ml2_odl.odl_features_retry_interval

    for times_tried in itertools.count():
        feature_configs = _fetch_features()
        if feature_configs is not None:
            break
        LOG.warning('Failed to retrieve ODL features, attempt %i', times_tried)
        time.sleep(wait_interval)


def has(feature):
    return feature in feature_configs


def get_config(feature):
    return feature_configs[feature]


def deinit():
    '''Set odl_features back to it's pre-initlialized '''
    global feature_configs
    feature_configs = copy.copy(EMPTY_FEATURES)


def _load_features(json):
    """parse and save features from json"""
    features = json['features']
    if 'feature' not in features:
        return copy.copy(EMPTY_FEATURES)

    # documentation on the JSON received can be found at:
    # https://github.com/opendaylight/neutron/blob/master/model/src/main/yang/neutron-extensions.yang
    LOG.info('Retrieved ODL features %s', features)
    response = {}
    for feature in features['feature']:
        cfg = feature.get('configuration', '')
        response[feature['service-provider-feature'].split(':')[1]] = cfg
    return response


def _fetch_features():
    '''Fetch the list of features declared by ODL.

    This function should be called once during initialization
    '''

    path = 'restconf/operational/neutron:neutron/neutron:features'
    features_url = utils.get_odl_url(path)

    client = odl_client.OpenDaylightRestClient.create_client(features_url)
    try:
        response = client.request('get')
    except exceptions.ConnectionError:
        LOG.error("Error connecting to ODL to retrieve features",
                  exc_info=True)
        return None

    if response.status_code == 400:
        LOG.debug('ODL does not support feature negotiation')
        return copy.copy(EMPTY_FEATURES)

    if response.status_code == 404:
        LOG.debug('No features configured')
        return copy.copy(EMPTY_FEATURES)

    if response.status_code != 200:
        LOG.warning('error fetching features: %i',
                    response.status_code)
        return None

    return _load_features(response.json())



#
# Copyright (C) 2017 Ericsson India Global Services Pvt Ltd.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

from neutron_lib.api.definitions import bgpvpn as bgpvpn_const
from neutron_lib.api.definitions import bgpvpn_vni as bgpvpn_vni_def
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from networking_bgpvpn.neutron.extensions import bgpvpn as bgpvpn_ext
from networking_bgpvpn.neutron.services.service_drivers import driver_api

from networking_odl.common import constants as odl_const
from networking_odl.common import odl_features
from networking_odl.common import postcommit
from networking_odl.journal import full_sync
from networking_odl.journal import journal


cfg.CONF.import_group('ml2_odl', 'networking_odl.common.config')

LOG = logging.getLogger(__name__)

BGPVPN_VNI = 'bgpvpn-vni'

BGPVPN_RESOURCES = {
    odl_const.ODL_BGPVPN: odl_const.ODL_BGPVPNS,
    odl_const.ODL_BGPVPN_NETWORK_ASSOCIATION:
        odl_const.ODL_BGPVPN_NETWORK_ASSOCIATIONS,

    odl_const.ODL_BGPVPN_ROUTER_ASSOCIATION:
        odl_const.ODL_BGPVPN_ROUTER_ASSOCIATIONS
}


@postcommit.add_postcommit('bgpvpn', 'net_assoc', 'router_assoc')
class OpenDaylightBgpvpnDriver(driver_api.BGPVPNDriver):

    """OpenDaylight BGPVPN Driver

    This code is the backend implementation for the OpenDaylight BGPVPN
    driver for Openstack Neutron.
    """

    @log_helpers.log_method_call
    def __init__(self, service_plugin):
        LOG.info("Initializing OpenDaylight BGPVPN v2 driver")
        super(OpenDaylightBgpvpnDriver, self).__init__(service_plugin)
        self.journal = journal.OpenDaylightJournalThread()
        full_sync.register(bgpvpn_const.ALIAS, BGPVPN_RESOURCES,
                           self.get_resources)
        if odl_features.has(BGPVPN_VNI):
            self.more_supported_extension_aliases = [bgpvpn_vni_def.ALIAS]

    @staticmethod
    def get_resources(context, resource_type):
        plugin = directory.get_plugin(bgpvpn_const.ALIAS)
        if resource_type == odl_const.ODL_BGPVPN:
            obj_getter = getattr(plugin,
                                 'get_%s' % BGPVPN_RESOURCES[resource_type])
            return obj_getter(context)

        method_name = 'get_%s' % BGPVPN_RESOURCES[resource_type]
        return full_sync.get_resources_require_id(plugin, context,
                                                  plugin.get_bgpvpns,
                                                  method_name)

    @log_helpers.log_method_call
    def create_bgpvpn_precommit(self, context, bgpvpn):
        journal.record(context, odl_const.ODL_BGPVPN,
                       bgpvpn['id'], odl_const.ODL_CREATE, bgpvpn)

    @log_helpers.log_method_call
    def update_bgpvpn_precommit(self, context, bgpvpn):
        journal.record(context, odl_const.ODL_BGPVPN,
                       bgpvpn['id'], odl_const.ODL_UPDATE, bgpvpn)

    @log_helpers.log_method_call
    def delete_bgpvpn_precommit(self, context, bgpvpn):
        journal.record(context, odl_const.ODL_BGPVPN,
                       bgpvpn['id'], odl_const.ODL_DELETE, [])

    @log_helpers.log_method_call
    def create_net_assoc_precommit(self, context, net_assoc):
        our_bgpvpn = None
        bgpvpns = self.get_bgpvpns(context)
        for bgpvpn in bgpvpns:
            # ODL only allows a network to be associated with one BGPVPN
            if bgpvpn['id'] == net_assoc['bgpvpn_id']:
                our_bgpvpn = bgpvpn
            else:
                if bgpvpn['networks'] and (net_assoc['network_id'] in
                                           bgpvpn['networks']):
                    raise bgpvpn_ext.BGPVPNNetworkAssocExistsAnotherBgpvpn(
                        driver="OpenDaylight V2",
                        network=net_assoc['network_id'],
                        bgpvpn=bgpvpn['id'])
        journal.record(context, odl_const.ODL_BGPVPN,
                       our_bgpvpn['id'], odl_const.ODL_UPDATE, our_bgpvpn)

    @log_helpers.log_method_call
    def delete_net_assoc_precommit(self, context, net_assoc):
        bgpvpn = self.get_bgpvpn(context, net_assoc['bgpvpn_id'])
        # NOTE(yamahata): precommit is called within db transaction.
        # so removing network_id is still associated.
        # it needs to be removed explicitly from dict.
        bgpvpn['networks'].remove(net_assoc['network_id'])
        journal.record(context, odl_const.ODL_BGPVPN,
                       bgpvpn['id'], odl_const.ODL_UPDATE, bgpvpn)

    @log_helpers.log_method_call
    def create_router_assoc_precommit(self, context, router_assoc):
        associated_routers = self.get_router_assocs(context,
                                                    router_assoc['bgpvpn_id'])
        for assoc_router in associated_routers:
            if(router_assoc["router_id"] != assoc_router["router_id"]):
                raise bgpvpn_ext.BGPVPNMultipleRouterAssocNotSupported(
                    driver="OpenDaylight V2")
        bgpvpn = self.get_bgpvpn(context, router_assoc['bgpvpn_id'])
        journal.record(context, odl_const.ODL_BGPVPN,
                       bgpvpn['id'], odl_const.ODL_UPDATE, bgpvpn)

    @log_helpers.log_method_call
    def delete_router_assoc_precommit(self, context, router_assoc):
        bgpvpn = self.get_bgpvpn(context, router_assoc['bgpvpn_id'])
        # NOTE(yamahata): precommit is called within db transaction.
        # so removing router_id is still associated.
        # it needs to be removed explicitly from dict.
        bgpvpn['routers'].remove(router_assoc['router_id'])
        journal.record(context, odl_const.ODL_BGPVPN,
                       bgpvpn['id'], odl_const.ODL_UPDATE, bgpvpn)



# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re
import ssl
import threading
import time

from oslo_config import cfg
from oslo_log import log
from oslo_serialization import jsonutils
from oslo_utils import excutils
from requests import codes
from requests import exceptions
import websocket

from networking_odl._i18n import _
from networking_odl.common import client as odl_client


cfg.CONF.import_group('ml2_odl', 'networking_odl.common.config')
LOG = log.getLogger(__name__)

ODL_OPERATIONAL_DATASTORE = "OPERATIONAL"
ODL_CONFIGURATION_DATASTORE = "CONFIGURATION"
ODL_NOTIFICATION_SCOPE_BASE = "BASE"
ODL_NOTIFICATION_SCOPE_ONE = "ONE"
ODL_NOTIFICATION_SCOPE_SUBTREE = "SUBTREE"

ODL_WEBSOCKET_DISCONNECTED = "ODL_WEBSOCKET_DISCONNECTED"
ODL_WEBSOCKET_CONNECTING = "ODL_WEBSOCKET_CONNECTING"
ODL_WEBSOCKET_CONNECTED = "ODL_WEBSOCKET_CONNECTED"


class OpenDaylightWebsocketClient(object):
    """Thread for the OpenDaylight Websocket """

    def __init__(self, odl_rest_client, path, datastore, scope, leaf_node_only,
                 packet_handler, timeout, status_cb=None):
        self.odl_rest_client = odl_rest_client
        self.path = path
        self.datastore = datastore
        self.scope = scope
        self.leaf_node_only = leaf_node_only
        self.packet_handler = packet_handler
        self.timeout = timeout
        self.exit_websocket_thread = False
        self.status_cb = status_cb
        self.current_status = ODL_WEBSOCKET_DISCONNECTED
        self._odl_sync_thread = self.start_odl_websocket_thread()

    @classmethod
    def odl_create_websocket(cls, odl_url, path, datastore, scope,
                             packet_handler, status_cb=None,
                             leaf_node_only=False):
        """Create a websocket connection with ODL.

                This method will create a websocket client based on path,
                datastore and scope params. On data recv from websocket
                packet_handler callback is called. status_cb callback can be
                provided if notifications are requried for socket status
                changes
        """

        if odl_url is None:
            LOG.error("invalid odl url", exc_info=True)
            raise ValueError(_("Invalid ODL URL"))

        odl_rest_client = odl_client.OpenDaylightRestClient.create_client(
            odl_url)
        return cls(
            odl_rest_client, path, datastore, scope, leaf_node_only,
            packet_handler, cfg.CONF.ml2_odl.timeout, status_cb
        )

    def start_odl_websocket_thread(self):
        # Start the websocket thread
        LOG.debug("starting a new websocket thread")
        odl_websocket_thread = threading.Thread(
            name='websocket',
            target=self.run_websocket_thread)
        odl_websocket_thread.start()
        return odl_websocket_thread

    def set_exit_flag(self, value=True):
        # set flag to exit
        self.exit_websocket_thread = value

    def run_websocket_thread(self, exit_after_run=False):
        # TBD connections are persistent so there is really no way to know
        # when it is a "first connection". We need to wait for the
        # dis/reconnect logic to be able to know this
        first_connection = True
        ws = None
        while not self.exit_websocket_thread:
            if exit_after_run:
                # Permanently waiting thread model breaks unit tests
                # Adding this arg to exit after one run for unit tests
                self.set_exit_flag()
            # connect if necessary
            if ws is None:
                try:
                    ws = self._connect_ws()
                except ValueError:
                    LOG.error("websocket irrecoverable error ")
                    return
                if ws is None:
                    time.sleep(cfg.CONF.ml2_odl.restconf_poll_interval)
                    continue
            # read off the websocket
            try:
                data = ws.recv()
                if not data:
                    LOG.warning("websocket received 0 bytes")
                    continue
            except websocket.WebSocketTimeoutException:
                continue
            except ssl.SSLError as e:
                message = e.args[0] if e.args else None
                # TODO(trozet): Workaround due to SSL Timeout not being caught
                # in websocket-client lib (issue 387).  Remove when fixed in
                # websocket-client lib.
                if message and 'timed out' in message:
                    continue

                LOG.error("SSL websocket unexpected exception, "
                          "closing and restarting...", exc_info=True)
                # TODO(rsood): Websocket reconnect can cause race
                # conditions
                self._close_ws(ws)
                ws = None
                continue
            except websocket.WebSocketConnectionClosedException:
                # per websocket-client, "If remote host closed the connection
                # or some network error happened"
                LOG.warning("websocket connection closed or IO error",
                            exc_info=True)
                self._close_ws(ws)
                ws = None
                continue
            except Exception:
                # Connection closed trigger reconnection
                LOG.error("websocket unexpected exception, "
                          "closing and restarting...", exc_info=True)
                # TODO(rsood): Websocket reconnect can cause race conditions
                self._close_ws(ws)
                ws = None
                continue

            # Call handler for data received
            try:
                self.packet_handler(data, first_connection)
                first_connection = False
            except Exception:
                LOG.error("Error in packet_handler callback",
                          exc_info=True)

        self._close_ws(ws)

    def _set_websocket_status(self, status):
        LOG.info("websocket transition to status %s", status)
        try:
            if self.status_cb:
                self.status_cb(status)
        except Exception:
            LOG.error("Error in status_cb", exc_info=True)

    def _subscribe_websocket(self):
        """ODL Websocket change notification subscription"""
        # Check ODL URL for details on this process
        # https://wiki.opendaylight.org/view/OpenDaylight_Controller:MD-SAL:Restconf:Change_event_notification_subscription#rpc_create-data-change-event-subscription # noqa: E501 # pylint: disable=line-too-long

        # Invoke rpc create-data-change-event-subscription
        ws_create_dce_subs_url = ("restconf/operations/sal-remote:"
                                  "create-data-change-event-subscription")
        odl_subscription_data = {'input': {
            'path': self.path,
            'sal-remote-augment:datastore': self.datastore,
            'sal-remote-augment:scope': self.scope,
            'sal-remote-augment:notification-output-type': 'JSON'
        }}
        try:
            response = self.odl_rest_client.sendjson('post',
                                                     ws_create_dce_subs_url,
                                                     odl_subscription_data)
            response.raise_for_status()
        except exceptions.ConnectionError:
            LOG.error("cannot connect to the opendaylight controller")
            return None
        except exceptions.HTTPError as e:
            # restconf returns 400 on operation when path is not available
            if e.response.status_code == codes.bad_request:
                LOG.debug("response code bad_request (400)"
                          "check path for websocket connection")
                raise ValueError(
                    _("bad_request (http400),check path.")) from e

            LOG.warning("websocket connection failed",
                        exc_info=True)
            return None
        except Exception:
            LOG.error("websocket subscription failed", exc_info=True)
            return None

        # Subscribing to stream. Returns websocket URL to listen to
        ws_dce_subs_url = """restconf/streams/stream/"""
        try:
            stream_name = response.json()
            stream_name = stream_name['output']['stream-name']
            url = ws_dce_subs_url + stream_name
            if self.leaf_node_only:
                url += "?odl-leaf-nodes-only=true"
            response = self.odl_rest_client.get(url)
            response.raise_for_status()
            stream_url = response.headers['location']
            LOG.debug("websocket stream URL: %s", stream_url)
            return stream_url
        except exceptions.ConnectionError:
            LOG.error("cannot connect to the opendaylight controller")
            return None
        except exceptions.HTTPError as e:
            # restconf returns 404 on operation when there is no entry
            if e.response.status_code == codes.not_found:
                LOG.debug("response code not_found (404)"
                          "unable to websocket connection url")
                raise ValueError(
                    _("bad_request (http400),check path")) from e

            LOG.warning("websocket connection failed")
            return None
        except ValueError:
            with excutils.save_and_reraise_exception():
                LOG.error("websocket subscribe got invalid stream name")
                return None
        except KeyError:
            LOG.error("websocket subscribe got bad stream data")
            raise ValueError(
                _("websocket subscribe bad stream data")) from None
        except Exception:
            LOG.error("websocket subscription failed", exc_info=True)
            return None

    def _socket_create_connection(self, stream_url):
        ws = None
        try:
            ws = websocket.create_connection(stream_url,
                                             timeout=self.timeout)
        except ValueError:
            with excutils.save_and_reraise_exception():
                LOG.error("websocket create connection invalid URL")
        except Exception:
            # Although a number of exceptions can occur here
            # we handle them all the same way, return None.
            # As such, enough to just "except Exception."
            LOG.exception("websocket create connection failed",
                          exc_info=True)
            return None
        if ws is None or not ws.connected:
            LOG.error("websocket create connection unsuccessful")
            return None

        LOG.debug("websocket connection established")
        return ws

    def _connect_ws(self):
        self._set_websocket_status(ODL_WEBSOCKET_CONNECTING)
        stream_url = self._subscribe_websocket()
        if stream_url is None:
            return None
        if 'https:' in self.odl_rest_client.url and 'wss:' not in stream_url:
            LOG.warning('TLS ODL URL detected, but websocket URL is not.  '
                        'Forcing websocket URL to TLS')
            stream_url = stream_url.replace('ws:', 'wss:')
        # Delay here causes websocket notification lose (ODL Bug 8299)
        ws = self._socket_create_connection(stream_url)
        if ws is not None:
            self._set_websocket_status(ODL_WEBSOCKET_CONNECTED)
        return ws

    def _close_ws(self, ws):
        LOG.debug("closing websocket")
        try:
            if ws is not None:
                ws.close()
        except Exception:
            LOG.error("Error while closing websocket", exc_info=True)
        self._set_websocket_status(ODL_WEBSOCKET_DISCONNECTED)


class EventDataParser(object):
    """Helper class to parse websocket notification data"""

    NOTIFICATION_TAG = 'notification'
    DC_NOTIFICATION_TAG = 'data-changed-notification'
    DC_EVENT_TAG = 'data-change-event'
    OPERATION_DELETE = 'deleted'
    OPERATION_CREATE = 'created'
    OPERATION_UPDATE = 'updated'

    def __init__(self, item):
        self.item = item

    @classmethod
    def get_item(cls, payload):
        try:
            data = jsonutils.loads(payload)
        except ValueError:
            LOG.warning("invalid websocket notification")
            return
        try:
            dn_events = (data[cls.NOTIFICATION_TAG]
                             [cls.DC_NOTIFICATION_TAG]
                             [cls.DC_EVENT_TAG])

            if not isinstance(dn_events, list):
                dn_events = [dn_events]

            for e in dn_events:
                yield cls(e)
        except KeyError:
            LOG.warning("invalid JSON for websocket notification")

    def get_fields(self):
        return (self.get_operation(),
                self.get_path(),
                self.get_data())

    def get_path(self):
        return self.item.get('path')

    def get_data(self):
        return self.item.get('data')

    def get_operation(self):
        return self.item.get('operation')

    @staticmethod
    def extract_field(text, key):
        pattern = r'\[' + key + r'=(.*?)\]'
        match = re.search(pattern, text)
        if match:
            return match.group(1)

        return None



# Copyright (c) 2017 Ericsson India Global Service Pvt Ltd.
# All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#
from neutron_lib.api.definitions import portbindings
from neutron_lib.services.trunk import constants as t_consts


SUPPORTED_INTERFACES = (
    portbindings.VIF_TYPE_OVS,
    portbindings.VIF_TYPE_VHOST_USER,
)

SUPPORTED_SEGMENTATION_TYPES = (
    t_consts.SEGMENTATION_TYPE_VLAN,
)



# Copyright (c) 2015 Intel Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_config import fixture as config_fixture

from networking_odl.common import client

from neutron.tests import base


class ClientTestCase(base.DietTestCase):
    def setUp(self):
        self.cfg = self.useFixture(config_fixture.Config())
        self.cfg.config(mechanism_drivers=[
                        'logger', 'opendaylight_v2'], group='ml2')
        super(ClientTestCase, self).setUp()

    def _set_config(self, url='http://127.0.0.1:9999', username='someuser',
                    password='somepass'):
        self.cfg.config(url=url, group='ml2_odl')
        self.cfg.config(username=username, group='ml2_odl')
        self.cfg.config(password=password, group='ml2_odl')

    def _test_missing_config(self, **kwargs):
        self._set_config(**kwargs)
        self.assertRaisesRegex(cfg.RequiredOptError,
                               r'value required for option \w+ in group '
                               r'\[ml2_odl\]',
                               client.OpenDaylightRestClient._check_opt,
                               cfg.CONF.ml2_odl.url)

    def test_valid_config(self):
        self._set_config()
        client.OpenDaylightRestClient._check_opt(cfg.CONF.ml2_odl.url)

    def test_missing_url_raises_exception(self):
        self._test_missing_config(url=None)

    def test_missing_username_raises_exception(self):
        self._test_missing_config(username=None)

    def test_missing_password_raises_exception(self):
        self._test_missing_config(password=None)



# Copyright 2017 Intel Corporation.
# Copyright 2017 Isaku Yamahata <isaku.yamahata at intel com>
#                               <isaku.yamahata at gmail com>
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re
import tokenize

from hacking.checks import docstrings
from hacking import core


_ND01_MSG = (
    "ND01: use OpenDaylight (capital D) instead of Opendaylight")  # noqa
_ND01_OPENDAYLIGHT = 'Opendaylight'  # noqa

_ND02_MSG = (
    "ND02: use the config fixture provided by oslo_config and use config()"
    " instead of %s")  # noqa

_ND02_REGEXP_DIRECT = re.compile(r'cfg\.CONF\..* =')

_ND03_MSG = (
    "ND03: The import of %s has a redundant alias."
)
_ND03_REGEXP_REDUNDANT_IMPORT_ALIAS = re.compile(r'.*import (.+) as \1$')


@core.flake8ext
def check_opendaylight_lowercase(logical_line, filename, noqa):
    """ND01 - Enforce using OpenDaylight."""
    if noqa:
        return

    if _ND01_OPENDAYLIGHT in logical_line:
        pos = logical_line.find(_ND01_OPENDAYLIGHT)
        yield (pos, _ND01_MSG)


def _check_opendaylight_lowercase(logical_line, tokens, noqa, token_type):
    """ND01 - Enforce using OpenDaylight in given token."""
    if noqa:
        return

    for _token_type, text, start_index, _, _ in tokens:
        if _token_type == token_type:
            pos = text.find(_ND01_OPENDAYLIGHT)
            if pos >= 0:
                msg = "{} in {}".format(
                    _ND01_MSG, tokenize.tok_name[token_type].lower())
                yield (start_index[1] + pos, msg)


@core.flake8ext
def check_opendaylight_lowercase_comment(logical_line, tokens, noqa):
    """ND01 - Enforce using OpenDaylight in comment."""

    for res in _check_opendaylight_lowercase(
            logical_line, tokens, noqa, tokenize.COMMENT):
        yield res


@core.flake8ext
def check_opendaylight_lowercase_string(logical_line, tokens, noqa):
    """ND01 - Enforce using OpenDaylight in string."""

    for res in _check_opendaylight_lowercase(
            logical_line, tokens, noqa, tokenize.STRING):
        yield res


@core.flake8ext
def check_opendaylight_lowercase_docstring(
        physical_line, previous_logical, tokens):
    """ND01 - Enforce using OpenDaylight in docstring."""
    docstring = docstrings.is_docstring(tokens, previous_logical)
    if docstring and _ND01_OPENDAYLIGHT in docstring:
        pos = physical_line.find(_ND01_OPENDAYLIGHT)
        return (pos, _ND01_MSG + " in docstring")
    return None


@core.flake8ext
def check_config_over_set_override(logical_line, filename, noqa):
    """ND02 - Enforcement of config fixture

    Enforce agreement of not use set_override() but use
    instead the fixture's config() helper for tests.
    """

    if noqa:
        return

    if 'networking_odl/tests/' not in filename:
        return

    if 'cfg.CONF.set_override' in logical_line:
        yield (0, _ND02_MSG % "using cfg.CONF.set_override()")


@core.flake8ext
def check_config_over_direct_override(logical_line, filename, noqa):
    """ND02 - Enforcement of config fixture

    Enforce usage of the fixture's config() helper instead
    of overriding a setting directly
    """

    if noqa:
        return

    if 'networking_odl/tests/' not in filename:
        return

    if _ND02_REGEXP_DIRECT.match(logical_line):
        yield (0, _ND02_MSG % "overriding it directly.")


@core.flake8ext
def check_redundant_import_alias(logical_line):
    """ND03 - Checking no redundant import alias.

    ND03: from neutron.plugins.ml2 import driver_context as driver_context
    OK: from neutron.plugins.ml2 import driver_context
    """

    match = re.match(_ND03_REGEXP_REDUNDANT_IMPORT_ALIAS, logical_line)
    if match:
        yield (0, _ND03_MSG % match.group(1))
